from __future__ import annotations

import calendar
import math
from string import whitespace
from typing import TYPE_CHECKING, Any

import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends.sql.compilers.base import (
    NULL,
    STAR,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import ClickHouseType
from xorq.vendor.ibis.backends.sql.dialects import ClickHouse


if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping


class ClickhouseAggGen(AggGen):
    def aggregate(self, compiler, name, *args, where=None, order_by=()):
        if order_by:
            raise com.UnsupportedOperationError(
                "ordering of order-sensitive aggregations via `order_by` is "
                "not supported for this backend"
            )
        # Clickhouse aggregate functions all have filtering variants with a
        # `If` suffix (e.g. `SumIf` instead of `Sum`).
        if where is not None:
            name += "If"
            args += (where,)
        return compiler.f[name](*args, dialect=compiler.dialect)


class ClickHouseCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = ClickHouse
    type_mapper = ClickHouseType

    agg = ClickhouseAggGen()

    supports_qualify = True

    UNSUPPORTED_OPS = (
        ops.RowID,
        ops.CumeDist,
        ops.PercentRank,
        ops.Time,
        ops.TimeDelta,
        ops.StringToTimestamp,
        ops.StringToDate,
        ops.Levenshtein,
    )

    SIMPLE_OPS = {
        ops.All: "min",
        ops.Any: "max",
        ops.ApproxCountDistinct: "uniqHLL12",
        ops.ApproxMedian: "median",
        ops.Arbitrary: "any",
        ops.ArgMax: "argMax",
        ops.ArgMin: "argMin",
        ops.ArrayContains: "has",
        ops.ArrayFlatten: "arrayFlatten",
        ops.ArrayIntersect: "arrayIntersect",
        ops.ArrayPosition: "indexOf",
        ops.BitwiseAnd: "bitAnd",
        ops.BitwiseNot: "bitNot",
        ops.BitwiseOr: "bitOr",
        ops.BitwiseXor: "bitXor",
        ops.Capitalize: "initcap",
        ops.CountDistinct: "uniq",
        ops.Date: "toDate",
        ops.E: "e",
        ops.EndsWith: "endsWith",
        ops.ExtractAuthority: "netloc",
        ops.ExtractDay: "toDayOfMonth",
        ops.ExtractDayOfYear: "toDayOfYear",
        ops.ExtractEpochSeconds: "toRelativeSecondNum",
        ops.ExtractFragment: "fragment",
        ops.ExtractHost: "domain",
        ops.ExtractHour: "toHour",
        ops.ExtractMinute: "toMinute",
        ops.ExtractMonth: "toMonth",
        ops.ExtractPath: "path",
        ops.ExtractProtocol: "protocol",
        ops.ExtractQuarter: "toQuarter",
        ops.ExtractSecond: "toSecond",
        ops.ExtractWeekOfYear: "toISOWeek",
        ops.ExtractYear: "toYear",
        ops.ExtractIsoYear: "toISOYear",
        ops.IntegerRange: "range",
        ops.IsInf: "isInfinite",
        ops.IsNan: "isNaN",
        ops.IsNull: "isNull",
        ops.Ln: "log",
        ops.Log10: "log10",
        ops.MapKeys: "mapKeys",
        ops.MapLength: "length",
        ops.MapMerge: "mapUpdate",
        ops.MapValues: "mapValues",
        ops.Median: "quantileExactExclusive",
        ops.NotNull: "isNotNull",
        ops.NullIf: "nullIf",
        ops.RegexReplace: "replaceRegexpAll",
        ops.RowNumber: "row_number",
        ops.StartsWith: "startsWith",
        ops.StrRight: "right",
        ops.Strftime: "formatDateTime",
        ops.StringLength: "length",
        ops.StringReplace: "replaceAll",
        ops.TimestampNow: "now",
        ops.TypeOf: "toTypeName",
        ops.Unnest: "arrayJoin",
    }

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_Cast(self, op, *, arg, to):
        _interval_cast_suffixes = {
            "s": "Second",
            "m": "Minute",
            "h": "Hour",
            "D": "Day",
            "W": "Week",
            "M": "Month",
            "Q": "Quarter",
            "Y": "Year",
        }

        if to.is_interval():
            suffix = _interval_cast_suffixes[to.unit.short]
            return self.f[f"toInterval{suffix}"](arg)

        result = self.cast(arg, to)
        if (timezone := getattr(to, "timezone", None)) is not None:
            return self.f.toTimeZone(result, timezone)
        return result

    def visit_TryCast(self, op, *, arg, to):
        return self.f.accurateCastOrNull(arg, self.type_mapper.to_string(to))

    def visit_ArrayIndex(self, op, *, arg, index):
        return arg[self.if_(index >= 0, index + 1, index)]

    def visit_ArrayRepeat(self, op, *, arg, times):
        param = sg.to_identifier("_")
        func = sge.Lambda(this=arg, expressions=[param])
        return self.f.arrayFlatten(self.f.arrayMap(func, self.f.range(times)))

    def visit_ArraySlice(self, op, *, arg, start, stop):
        start = self._add_parens(op.start, start)
        start_correct = self.if_(start < 0, start, start + 1)

        if stop is not None:
            stop = self._add_parens(op.stop, stop)

            length = self.if_(
                stop < 0,
                stop,
                self.if_(
                    start < 0,
                    self.f.greatest(0, stop - (self.f.length(arg) + start)),
                    self.f.greatest(0, stop - start),
                ),
            )
            return self.f.arraySlice(arg, start_correct, length)
        else:
            return self.f.arraySlice(arg, start_correct)

    def visit_CountStar(self, op, *, where, arg):
        if where is not None:
            return self.f.countIf(where)
        return sge.Count(this=STAR)

    def _visit_quantile(self, func, arg, quantile, where):
        return sge.ParameterizedAgg(
            this=f"{func}If" if where is not None else func,
            expressions=util.promote_list(quantile),
            params=[arg, where] if where is not None else [arg],
        )

    def visit_Quantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantile", arg, quantile, where)

    def visit_MultiQuantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantiles", arg, quantile, where)

    def visit_ApproxQuantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantileTDigest", arg, quantile, where)

    def visit_ApproxMultiQuantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantilesTDigest", arg, quantile, where)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "pop":
            raise ValueError(
                "ClickHouse only implements `sample` correlation coefficient"
            )
        return self.agg.corr(left, right, where=where)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise com.UnsupportedOperationError(
                "String find doesn't support end argument"
            )

        if start is not None:
            return self.f.position(arg, substr, start)

        return self.f.position(arg, substr)

    def visit_RegexSearch(self, op, *, arg, pattern):
        return sge.RegexpLike(this=arg, expression=pattern)

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        arg = self.cast(arg, dt.String(nullable=False))

        pattern = self.f.concat("(", pattern, ")")

        if index is None:
            index = 0

        index += 1

        then = self.f.extractGroups(arg, pattern)[index]

        return self.if_(self.f.notEmpty(then), then, NULL)

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.indexOf(self.f.array(*values), needle)

    def visit_Sign(self, op, *, arg):
        """Workaround for missing sign function in older versions of clickhouse."""
        return self.f.intDivOrZero(arg, self.f.abs(arg))

    def visit_Hash(self, op, *, arg):
        return self.f.reinterpretAsInt64(self.f.sipHash64(arg))

    def visit_HashBytes(self, op, *, arg, how):
        supported_algorithms = {
            "md5": "MD5",
            "MD5": "MD5",
            "halfMD5": "halfMD5",
            "SHA1": "SHA1",
            "sha1": "SHA1",
            "SHA224": "SHA224",
            "sha224": "SHA224",
            "SHA256": "SHA256",
            "sha256": "SHA256",
            "intHash32": "intHash32",
            "intHash64": "intHash64",
            "cityHash64": "cityHash64",
            "sipHash64": "sipHash64",
            "sipHash128": "sipHash128",
        }
        if (funcname := supported_algorithms.get(how)) is None:
            raise com.UnsupportedOperationError(f"Unsupported hash algorithm {how}")

        return self.f[funcname](arg)

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        dtype = op.dtype
        if dtype.unit.short in ("ms", "us", "ns"):
            raise com.UnsupportedOperationError(
                "Clickhouse doesn't support subsecond interval resolutions"
            )
        return super().visit_IntervalFromInteger(op, arg=arg, unit=unit)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_inet():
            v = str(value)
            return self.f.toIPv6(v) if ":" in v else self.f.toIPv4(v)
        elif dtype.is_decimal():
            precision = dtype.precision
            if precision is None or not 1 <= precision <= 76:
                raise NotImplementedError(
                    f"Unsupported precision. Supported values: [1 : 76]. Current value: {precision!r}"
                )

            if 1 <= precision <= 9:
                type_name = self.f.toDecimal32
            elif 10 <= precision <= 18:
                type_name = self.f.toDecimal64
            elif 19 <= precision <= 38:
                type_name = self.f.toDecimal128
            else:
                type_name = self.f.toDecimal256
            return type_name(value, dtype.scale)
        elif dtype.is_numeric():
            if not math.isfinite(value):
                return sge.Literal.number(str(value))
            return sge.convert(value)
        elif dtype.is_interval():
            if dtype.unit.short in ("ms", "us", "ns"):
                raise com.UnsupportedOperationError(
                    "Clickhouse doesn't support subsecond interval resolutions"
                )

            return sge.Interval(
                this=sge.convert(str(value)), unit=dtype.resolution.upper()
            )
        elif dtype.is_timestamp():
            funcname = "parseDateTime"

            if micros := value.microsecond:
                funcname += "64"

            funcname += "BestEffort"

            args = [value.isoformat()]

            if micros % 1000:
                args.append(6)
            elif micros // 1000:
                args.append(3)

            if (timezone := dtype.timezone) is not None:
                args.append(timezone)

            return self.f[funcname](*args)
        elif dtype.is_date():
            return self.f.toDate(value.isoformat())
        elif dtype.is_array():
            value_type = dtype.value_type
            values = [
                self.visit_Literal(
                    ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                )
                for v in value
            ]
            return self.f.array(*values)
        elif dtype.is_map():
            value_type = dtype.value_type
            keys = []
            values = []

            for k, v in value.items():
                keys.append(sge.convert(k))
                values.append(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type),
                        value=v,
                        dtype=value_type,
                    )
                )

            return self.f.map(self.f.array(*keys), self.f.array(*values))
        elif dtype.is_struct():
            fields = [
                self.visit_Literal(
                    ops.Literal(v, dtype=field_type), value=v, dtype=field_type
                )
                for field_type, v in zip(dtype.types, value.values())
            ]
            return self.f.tuple(*fields)
        else:
            return None

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        if (unit := unit.short) in {"ms", "us", "ns"}:
            raise com.UnsupportedOperationError(f"{unit!r} unit is not supported!")
        return self.f.toDateTime(arg)

    def visit_TimestampTruncate(self, op, *, arg, unit):
        if (short := unit.short) == "W":
            func = "toMonday"
        else:
            func = f"toStartOf{unit.singular.capitalize()}"

        if short in ("s", "ms", "us", "ns"):
            arg = self.f.toDateTime64(arg, op.arg.dtype.scale or 0)
        return self.f[func](arg)

    visit_TimeTruncate = visit_TimestampTruncate

    def visit_DateTruncate(self, op, *, arg, unit):
        if unit.short == "W":
            func = "toMonday"
        else:
            func = f"toStartOf{unit.singular.capitalize()}"
        return self.f[func](arg)

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        if offset is not None:
            raise com.UnsupportedOperationError(
                "Timestamp bucket with offset is not supported"
            )

        return self.f.toStartOfInterval(arg, interval)

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.f.toDate(
            self.f.concat(
                self.f.toString(year),
                "-",
                self.f.leftPad(self.f.toString(month), 2, "0"),
                "-",
                self.f.leftPad(self.f.toString(day), 2, "0"),
            )
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds, **_
    ):
        to_datetime = self.f.toDateTime(
            self.f.concat(
                self.f.toString(year),
                "-",
                self.f.leftPad(self.f.toString(month), 2, "0"),
                "-",
                self.f.leftPad(self.f.toString(day), 2, "0"),
                " ",
                self.f.leftPad(self.f.toString(hours), 2, "0"),
                ":",
                self.f.leftPad(self.f.toString(minutes), 2, "0"),
                ":",
                self.f.leftPad(self.f.toString(seconds), 2, "0"),
            )
        )
        if timezone := op.dtype.timezone:
            return self.f.toTimeZone(to_datetime, timezone)
        return to_datetime

    def visit_StringSplit(self, op, *, arg, delimiter):
        return self.f.splitByString(
            delimiter, self.cast(arg, dt.String(nullable=False))
        )

    def visit_GroupConcat(self, op, *, arg, sep, where, order_by):
        if order_by:
            raise com.UnsupportedOperationError(
                "ordering of order-sensitive aggregations via `order_by` is "
                "not supported for this backend"
            )
        call = self.agg.groupArray(arg, where=where)
        return self.if_(self.f.empty(call), NULL, self.f.arrayStringConcat(call, sep))

    def visit_Cot(self, op, *, arg):
        return 1.0 / self.f.tan(arg)

    def visit_StructColumn(self, op, *, values, **_):
        return self.f.tuple(*values)

    def visit_Clip(self, op, *, arg, lower, upper):
        if upper is not None:
            arg = self.if_(self.f.isNull(arg), NULL, self.f.least(upper, arg))

        if lower is not None:
            arg = self.if_(self.f.isNull(arg), NULL, self.f.greatest(lower, arg))

        return arg

    def visit_StructField(self, op, *, arg, field: str):
        arg_dtype = op.arg.dtype
        idx = arg_dtype.names.index(field)
        return self.cast(sge.Dot(this=arg, expression=sge.convert(idx + 1)), op.dtype)

    def visit_Repeat(self, op, *, arg, times):
        return self.f.repeat(arg, self.f.accurateCast(times, "UInt64"))

    def visit_StringContains(self, op, haystack, needle):
        return self.f.position(haystack, needle) > 0

    def visit_Strip(self, op, *, arg):
        return sge.Trim(
            this=arg, position="BOTH", expression=sge.Literal.string(whitespace)
        )

    def visit_DayOfWeekIndex(self, op, *, arg):
        weekdays = len(calendar.day_name)
        return (((self.f.toDayOfWeek(arg) - 1) % weekdays) + weekdays) % weekdays

    def visit_DayOfWeekName(self, op, *, arg):
        # ClickHouse 20 doesn't support dateName
        #
        # ClickHouse 21 supports dateName is broken for regexen:
        # https://github.com/ClickHouse/ClickHouse/issues/32777
        #
        # ClickHouses 20 and 21 also have a broken case statement hence the ifnull:
        # https://github.com/ClickHouse/ClickHouse/issues/32849
        #
        # We test against 20 in CI, so we implement day_of_week_name as follows
        days = calendar.day_name
        num_weekdays = len(days)
        base = (
            ((self.f.toDayOfWeek(arg) - 1) % num_weekdays) + num_weekdays
        ) % num_weekdays
        return sge.Case(
            this=base,
            ifs=list(map(self.if_, *zip(*enumerate(days)))),
            default=sge.convert(""),
        )

    def visit_Map(self, op, *, keys, values):
        # cast here to allow lookups of nullable columns
        return self.cast(self.f.tuple(keys, values), op.dtype)

    def visit_MapGet(self, op, *, arg, key, default):
        return self.if_(self.f.mapContains(arg, key), arg[key], default)

    def visit_ArrayConcat(self, op, *, arg):
        return self.f.arrayConcat(*arg)

    def visit_BitAndOrXor(self, op, *, arg, where):
        if not (dtype := op.arg.dtype).is_unsigned_integer():
            nbits = dtype.nbytes * 8
            arg = self.f[f"reinterpretAsUInt{nbits}"](arg)
        return self.agg[f"group{type(op).__name__}"](arg, where=where)

    visit_BitAnd = visit_BitOr = visit_BitXor = visit_BitAndOrXor

    def visit_StandardDevVariance(self, op, *, how, where, **kw):
        funcs = {
            ops.StandardDev: "stddev",
            ops.Variance: "var",
            ops.Covariance: "covar",
        }
        func = funcs[type(op)]
        variants = {"sample": f"{func}Samp", "pop": f"{func}Pop"}
        funcname = variants[how]
        return self.agg[funcname](*kw.values(), where=where)

    visit_StandardDev = visit_Variance = visit_Covariance = visit_StandardDevVariance

    def visit_ArrayDistinct(self, op, *, arg):
        null_element = self.if_(
            self.f.countEqual(arg, NULL) > 0, self.f.array(NULL), self.f.array()
        )
        return self.f.arrayConcat(self.f.arrayDistinct(arg), null_element)

    def visit_ExtractMicrosecond(self, op, *, arg):
        dtype = op.dtype
        return self.cast(
            self.f.toUnixTimestamp64Micro(self.cast(arg, op.arg.dtype.copy(scale=6)))
            % 1_000_000,
            dtype,
        )

    def visit_ExtractMillisecond(self, op, *, arg):
        dtype = op.dtype
        return self.cast(
            self.f.toUnixTimestamp64Milli(self.cast(arg, op.arg.dtype.copy(scale=3)))
            % 1_000,
            dtype,
        )

    def visit_LagLead(self, op, *, arg, offset, default):
        args = [arg]

        if default is not None:
            if offset is None:
                offset = 1

            args.append(offset)
            args.append(default)
        elif offset is not None:
            args.append(offset)

        func = self.f[f"{type(op).__name__.lower()}InFrame"]
        return func(*args)

    visit_Lag = visit_Lead = visit_LagLead

    def visit_ExtractFile(self, op, *, arg):
        return self.f.cutFragment(self.f.pathFull(arg))

    def visit_ExtractQuery(self, op, *, arg, key):
        if key is not None:
            return self.f.extractURLParameter(arg, key)
        else:
            return self.f.queryString(arg)

    def visit_ArrayStringJoin(self, op, *, arg, sep):
        return self.f.arrayStringConcat(arg, sep)

    def visit_ArrayMap(self, op, *, arg, param, body):
        func = sge.Lambda(this=body, expressions=[param])
        return self.f.arrayMap(func, arg)

    def visit_ArrayFilter(self, op, *, arg, param, body):
        func = sge.Lambda(this=body, expressions=[param])
        return self.f.arrayFilter(func, arg)

    def visit_ArrayRemove(self, op, *, arg, other):
        x = sg.to_identifier(util.gen_name("x"))
        should_keep_null = sg.and_(x.is_(NULL), other.is_(sg.not_(NULL)))
        cond = sg.or_(x.neq(other), should_keep_null)
        return self.f.arrayFilter(sge.Lambda(this=cond, expressions=[x]), arg)

    def visit_ArrayUnion(self, op, *, left, right):
        arg = self.f.arrayConcat(left, right)
        null_element = self.if_(
            self.f.countEqual(arg, NULL) > 0, self.f.array(NULL), self.f.array()
        )
        return self.f.arrayConcat(self.f.arrayDistinct(arg), null_element)

    def visit_ArrayZip(self, op: ops.ArrayZip, *, arg, **_: Any) -> str:
        return self.f.arrayZip(*arg)

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the clickhouse backend"
            )
        return self.agg.groupArray(arg, where=where, order_by=order_by)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the clickhouse backend"
            )
        return self.agg.any(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the clickhouse backend"
            )
        return self.agg.anyLast(arg, where=where, order_by=order_by)

    def visit_CountDistinctStar(
        self, op: ops.CountDistinctStar, *, where, **_: Any
    ) -> str:
        columns = self.f.tuple(*map(sg.column, op.arg.schema.names))

        if where is not None:
            return self.f.countDistinctIf(columns, where)
        else:
            return self.f.countDistinct(columns)

    def visit_TimestampRange(self, op, *, start, stop, step):
        unit = op.step.dtype.unit.name.lower()

        if not isinstance(op.step, ops.Literal):
            raise com.UnsupportedOperationError(
                "ClickHouse doesn't support non-literal step values"
            )

        step_value = op.step.value

        offset = sg.to_identifier("offset")

        func = sge.Lambda(
            this=self.f.dateAdd(sg.to_identifier(unit), offset, start),
            expressions=[offset],
        )

        if step_value == 0:
            return self.f.array()

        return self.f.arrayMap(
            func, self.f.range(0, self.f.timestampDiff(unit, start, stop), step_value)
        )

    def visit_RegexSplit(self, op, *, arg, pattern):
        return self.f.splitByRegexp(pattern, self.cast(arg, dt.String(nullable=False)))

    def visit_RandomScalar(self, op, **kwargs):
        return self.f.randCanonical()

    def visit_RandomUUID(self, op, **kwargs):
        return self.f.generateUUIDv4()

    @staticmethod
    def _generate_groups(groups):
        return groups

    def visit_DropColumns(self, op, *, parent, columns_to_drop):
        quoted = self.quoted
        excludes = [sg.column(column, quoted=quoted) for column in columns_to_drop]
        star = sge.Star(**{"except": excludes})
        table = sg.to_identifier(parent.alias_or_name, quoted=quoted)
        column = sge.Column(this=star, table=table)
        return sg.select(column).from_(parent)

    def visit_TableUnnest(
        self,
        op,
        *,
        parent,
        column,
        column_name: str,
        offset: str | None,
        keep_empty: bool,
    ):
        quoted = self.quoted

        column_alias = sg.to_identifier(
            util.gen_name("table_unnest_column"), quoted=quoted
        )

        table = sg.to_identifier(parent.alias_or_name, quoted=quoted)

        selcols = []

        overlaps_with_parent = column_name in op.parent.schema
        computed_column = column_alias.as_(column_name, quoted=quoted)

        if offset is not None:
            if overlaps_with_parent:
                selcols.append(
                    sge.Column(this=sge.Star(replace=[computed_column]), table=table)
                )
            else:
                selcols.append(sge.Column(this=STAR, table=table))
                selcols.append(computed_column)

            offset = sg.to_identifier(offset, quoted=quoted)
            selcols.append(offset)
        elif overlaps_with_parent:
            selcols.append(
                sge.Column(this=sge.Star(replace=[computed_column]), table=table)
            )
        else:
            selcols.append(sge.Column(this=STAR, table=table))
            selcols.append(computed_column)

        select = (
            sg.select(*selcols)
            .from_(parent)
            .join(
                sge.Join(
                    this=column.as_(column_alias, quoted=quoted),
                    kind="ARRAY",
                    side=None if not keep_empty else "LEFT",
                )
            )
        )

        if offset is not None:
            param = sg.to_identifier(util.gen_name("arr_enum"))
            func = sge.Lambda(this=param - 1, expressions=[param])
            return select.join(
                self.f.arrayMap(func, self.f.arrayEnumerate(column)).as_(offset)
            )

        return select

    def _cleanup_names(
        self, exprs: Mapping[str, sge.Expression]
    ) -> Iterator[sge.Expression]:
        """Compose `_gen_valid_name` and `_dedup_name` to clean up names in projections.

        ClickHouse has a bug where this fails to find the final `"o"."a"` column:

        ```sql
        SELECT
          "o"."a"
        FROM (
          SELECT
            "w"."a"
          FROM "t" AS "s"
          INNER JOIN "t" AS "w"
          USING ("a")
        ) AS "o"
        ```

        Adding a redundant aliasing operation (`"w"."a" AS "a"`) helps
        ClickHouse.
        """
        quoted = self.quoted
        return (
            value.as_(self._gen_valid_name(name), quoted=quoted, copy=False)
            for name, value in exprs.items()
        )

    def _array_reduction(self, arg):
        x = sg.to_identifier("x", quoted=self.quoted)
        not_null = sge.Lambda(this=x.is_(sg.not_(NULL)), expressions=[x])
        return self.f.arrayFilter(not_null, arg)

    def visit_ArrayMin(self, op, *, arg):
        return self.f.arrayReduce("min", self._array_reduction(arg))

    visit_ArrayAll = visit_ArrayMin

    def visit_ArrayMax(self, op, *, arg):
        return self.f.arrayReduce("max", self._array_reduction(arg))

    visit_ArrayAny = visit_ArrayMax

    def visit_ArraySum(self, op, *, arg):
        return self.f.arrayReduce("sum", self._array_reduction(arg))

    def visit_ArrayMean(self, op, *, arg):
        return self.f.arrayReduce("avg", self._array_reduction(arg))

    def _promote_bitshift_inputs(self, *, op, left, right):
        # clickhouse is incredibly pedantic about types allowed in bit shifting
        #
        # e.g., a UInt8 cannot be bitshift by more than 8 bits, UInt16 by more
        # than 16, and so on.
        #
        # This is why something like Ibis is necessary so that people have just
        # _consistent_ things, let alone *nice* things.
        left_dtype = op.left.dtype
        right_dtype = op.right.dtype

        if left_dtype != right_dtype:
            promoted = dt.higher_precedence(left_dtype, right_dtype)
            return self.cast(left, promoted), self.cast(right, promoted)
        return left, right

    def visit_BitwiseLeftShift(self, op, *, left, right):
        return self.f.bitShiftLeft(
            *self._promote_bitshift_inputs(op=op, left=left, right=right)
        )

    def visit_BitwiseRightShift(self, op, *, left, right):
        return self.f.bitShiftRight(
            *self._promote_bitshift_inputs(op=op, left=left, right=right)
        )

    def visit_MapContains(self, op, *, arg, key):
        return self.if_(
            sg.or_(arg.is_(NULL), key.is_(NULL)), NULL, self.f.mapContains(arg, key)
        )


compiler = ClickHouseCompiler()



from __future__ import annotations

from functools import partial
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import pyarrow as pa
import toolz

import xorq.vendor.ibis.expr.types as ir
from xorq.backends.xorq import connect as xo_connect
from xorq.common.utils.inspect_utils import (
    get_arguments,
)
from xorq.config import _backend_init
from xorq.expr.relations import (
    Read,
)
from xorq.vendor import ibis
from xorq.vendor.ibis import Schema
from xorq.vendor.ibis.util import (
    gen_name,
    normalize_filenames,
)


if TYPE_CHECKING:
    from xorq.backends.xorq import Backend


DEFAULT_CHUNKSIZE = 10_000


def make_read_kwargs(f, *args, **kwargs):
    # FIXME: if any kwarg is a dictionary, we'll fail Concrete's hashable requirement, so just pickle
    read_kwargs = get_arguments(f, *args, **kwargs)
    kwargs = read_kwargs.pop("kwargs", {})
    tpl = tuple(read_kwargs.items()) + tuple(kwargs.items())
    return tpl


def normalize_read_path_stat(path):
    stat = path.stat()
    tpls = tuple(
        (attrname, getattr(stat, attrname))
        for attrname in (
            "st_mtime",
            "st_size",
            # mtime, size <?-?> md5sum
            "st_ino",
        )
    )
    return tpls


@toolz.curry
def infer_csv_schema_pandas(path, chunksize=DEFAULT_CHUNKSIZE, **kwargs):
    import pandas as pd

    path = normalize_filenames(path)
    gen = pd.read_csv(path[0], chunksize=chunksize, **kwargs)
    df = next(gen)
    batch = pa.RecordBatch.from_pandas(df, preserve_index=False)
    schema = ibis.Schema.from_pyarrow(batch.schema)
    return schema


def read_csv_rbr(*args, schema=None, chunksize=DEFAULT_CHUNKSIZE, dtype=None, **kwargs):
    """Deferred and streaming csv reading via pandas"""
    import pandas as pd

    if dtype is not None:
        raise Exception("pass `dtype` as pyarrow `schema`")
    if chunksize is None:
        raise ValueError("chunksize must not be `None`")
    if schema is not None:
        schema = ibis.schema(schema)
        dtype = {col: typ.to_pandas() for col, typ in schema.items()}
        schema = schema.to_pyarrow()
    # schema is always nullable (this is good)
    paths = normalize_filenames(*args)

    gen = map(
        partial(pa.RecordBatch.from_pandas, preserve_index=False),
        chain.from_iterable(
            pd.read_csv(
                path,
                dtype=dtype,
                chunksize=chunksize,
                **kwargs,
            )
            for path in paths
        ),
    )
    if schema is None:
        (el, gen) = toolz.peek(gen)
        schema = el.schema

    def cast_gen():
        yield from (batch.cast(schema) for batch in gen)

    rbr = pa.RecordBatchReader.from_batches(
        schema,
        cast_gen(),
    )
    return rbr


def deferred_read_csv(
    path: str | Path,
    con: Backend | None = None,
    table_name: str | None = None,
    schema: Schema | None = None,
    normalize_method: Callable = normalize_read_path_stat,
    **kwargs,
) -> ir.Table:
    """
    Create a deferred read operation for CSV files that will execute only when needed.

    This function creates a representation of a read operation that doesn't immediately
    load data into memory. Instead, it registers the operation to be performed when
    the resulting expression is executed.

    The function works with different backend engines (pandas, duckdb, postgres, etc.)
    and adapts the read parameters accordingly.


    Parameters
    ----------
    con : Backend
        The connection object representing the backend where the CSV will be read.
        This can be any backend that supports reading CSV files (pandas, duckdb,
        postgres, etc.).

    path : str or Path
        The path to the CSV file to be read. This can be a local file path or a URL.

    table_name : str, optional
        The name to give to the resulting table in the backend. If not provided,
        a unique name will be generated automatically.

    schema : Schema, optional
        The schema definition for the CSV data. If not provided, the schema will
        be inferred from the data by sampling the CSV file.

    kwargs : Any
        Additional keyword arguments that will be passed to the backend's read_csv
        method.

    Returns
    -------
    Expr
        An expression representing the deferred read operation.
    """

    infer_schema = kwargs.pop("infer_schema", infer_csv_schema_pandas)
    deferred_read_csv.method_name = method_name = "read_csv"

    if con is None:
        con = _backend_init()

    method = getattr(con, method_name)

    if table_name is None:
        table_name = gen_name(f"xorq-{method_name}")
    if schema is None:
        schema = infer_schema(path)
    if con.name == "pandas":
        # FIXME: determine how to best handle schema
        read_kwargs = make_read_kwargs(method, path, table_name, **kwargs)
    elif con.name == "duckdb":
        read_kwargs = make_read_kwargs(
            method, path, table_name=table_name, columns=schema, **kwargs
        )
    else:
        read_kwargs = make_read_kwargs(
            method, path, table_name, schema=schema, **kwargs
        )
    return Read(
        method_name=method_name,
        name=table_name,
        schema=schema,
        source=con,
        read_kwargs=read_kwargs,
        normalize_method=normalize_method,
    ).to_expr()


def deferred_read_parquet(
    path: str | Path,
    con: Backend | None = None,
    table_name: str | None = None,
    schema: Schema | None = None,
    normalize_method: Callable = normalize_read_path_stat,
    **kwargs,
) -> ir.Table:
    """
     Create a deferred read operation for Parquet files that will execute only when needed.

    This function creates a representation of a read operation that doesn't immediately
    load data into memory. Instead, it registers the operation to be performed when
    the resulting expression is executed.

    Parameters
    ----------
    con : Backend
        The connection object representing the backend where the Parquet data will be read.

    path : str or Path
        The path to the Parquet file or directory to be read.

    table_name : str, optional
        The name to give to the resulting table in the backend. If not provided,
        a unique name will be generated automatically.

    normalize_method : Callable, optional
     The method that returns the values to be used in the hashing of the Read operation.

    **kwargs : dict
        Additional keyword arguments passed to the backend's read_parquet method.

     Returns
     -------
     Expr
         An expression representing the deferred read operation.
    """

    deferred_read_parquet.method_name = method_name = "read_parquet"
    if con is None:
        con = _backend_init()
    method = getattr(con, method_name)
    if table_name is None:
        table_name = gen_name(f"xorq-{method_name}")
    schema = schema or xo_connect().read_parquet(path).schema()
    read_kwargs = make_read_kwargs(method, path, table_name=table_name, **kwargs)
    return Read(
        method_name=method_name,
        name=table_name,
        schema=schema,
        source=con,
        read_kwargs=read_kwargs,
        normalize_method=normalize_method,
    ).to_expr()


def rbr_wrapper(reader, clean_up):
    def gen():
        yield from reader
        clean_up()

    return pa.RecordBatchReader.from_batches(reader.schema, gen())



import os
from pathlib import Path

import pytest

import xorq.api as xo
from xorq.common.utils.aws_utils import (
    connection_is_set,
    make_s3_credentials_defaults,
)
from xorq.conftest import array_types_df
from xorq.vendor import ibis


FIXTURES_DIR = Path(__file__).parent / "fixtures"

TEST_TABLES = {
    "functional_alltypes": ibis.schema(
        {
            "id": "int32",
            "bool_col": "boolean",
            "tinyint_col": "int8",
            "smallint_col": "int16",
            "int_col": "int32",
            "bigint_col": "int64",
            "float_col": "float32",
            "double_col": "float64",
            "date_string_col": "string",
            "string_col": "string",
            "timestamp_col": "timestamp",
            "year": "int32",
            "month": "int32",
        }
    ),
    "diamonds": ibis.schema(
        {
            "carat": "float64",
            "cut": "string",
            "color": "string",
            "clarity": "string",
            "depth": "float64",
            "table": "float64",
            "price": "int64",
            "x": "float64",
            "y": "float64",
            "z": "float64",
        }
    ),
    "batting": ibis.schema(
        {
            "playerID": "string",
            "yearID": "int64",
            "stint": "int64",
            "teamID": "string",
            "lgID": "string",
            "G": "int64",
            "AB": "int64",
            "R": "int64",
            "H": "int64",
            "X2B": "int64",
            "X3B": "int64",
            "HR": "int64",
            "RBI": "int64",
            "SB": "int64",
            "CS": "int64",
            "BB": "int64",
            "SO": "int64",
            "IBB": "int64",
            "HBP": "int64",
            "SH": "int64",
            "SF": "int64",
            "GIDP": "int64",
        }
    ),
    "awards_players": ibis.schema(
        {
            "playerID": "string",
            "awardID": "string",
            "yearID": "int64",
            "lgID": "string",
            "tie": "string",
            "notes": "string",
        }
    ),
    "astronauts": ibis.schema(
        {
            "id": "int64",
            "number": "int64",
            "nationwide_number": "int64",
            "name": "string",
            "original_name": "string",
            "sex": "string",
            "year_of_birth": "int64",
            "nationality": "string",
            "military_civilian": "string",
            "selection": "string",
            "year_of_selection": "int64",
            "mission_number": "int64",
            "total_number_of_missions": "int64",
            "occupation": "string",
            "year_of_mission": "int64",
            "mission_title": "string",
            "ascend_shuttle": "string",
            "in_orbit": "string",
            "descend_shuttle": "string",
            "hours_mission": "float64",
            "total_hrs_sum": "float64",
            "field21": "int64",
            "eva_hrs_mission": "float64",
            "total_eva_hrs": "float64",
        }
    ),
}

have_s3_credentials = connection_is_set(make_s3_credentials_defaults())


def pytest_runtest_setup(item):
    if any(mark.name == "s3" for mark in item.iter_markers()):
        if not have_s3_credentials:
            pytest.skip("cannot run s3 tests without s3 credentials")
    if any(mark.name == "gcs" for mark in item.iter_markers()):
        if not os.environ.get("GCS_ENABLED", False):
            pytest.skip("GCS are not enabled by default")


@pytest.fixture(scope="session")
def ddl_file(root_dir):
    ddl_dir = root_dir / "db" / "datafusion.sql"
    return ddl_dir


def statements(ddl_file: Path):
    return (
        statement
        for statement in map(str.strip, ddl_file.read_text().split(";"))
        if statement
    )


@pytest.fixture(scope="session")
def con(data_dir, ddl_file):
    conn = xo.connect()
    parquet_dir = data_dir / "parquet"
    conn.read_parquet(
        parquet_dir / "functional_alltypes.parquet", "functional_alltypes"
    )
    conn.read_parquet(parquet_dir / "batting.parquet", "batting")
    conn.read_parquet(parquet_dir / "diamonds.parquet", "diamonds")
    conn.read_parquet(parquet_dir / "astronauts.parquet", "astronauts")
    conn.read_parquet(parquet_dir / "awards_players.parquet", "awards_players")

    conn.create_table("array_types", array_types_df)

    if ddl_file.is_file() and ddl_file.name.endswith(".sql"):
        for statement in statements(ddl_file):
            with conn._safe_raw_sql(statement):  # noqa
                pass

    return conn


@pytest.fixture(scope="session")
def functional_alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def batting(con):
    return con.table("batting")


@pytest.fixture(scope="session")
def batting_df(batting):
    return batting.execute()


@pytest.fixture(scope="session")
def awards_players(con):
    return con.table("awards_players")


@pytest.fixture(scope="session")
def awards_players_df(awards_players):
    return awards_players.execute(limit=None)


@pytest.fixture(scope="session")
def sorted_df(df):
    return df.sort_values("id").reset_index(drop=True)


@pytest.fixture(scope="session")
def diamonds(con):
    return con.table("diamonds")


@pytest.fixture(scope="session")
def array_types(con):
    return con.table("array_types")


@pytest.fixture(scope="session")
def struct(con):
    return con.table("structs")


@pytest.fixture(scope="session")
def struct_df(struct):
    return struct.execute()



import pandas as pd
import snowflake.connector
from adbc_driver_snowflake import (
    DatabaseOptions,
    dbapi,
)
from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)

from xorq.backends.snowflake import Backend as SnowflakeBackend
from xorq.backends.snowflake import SnowflakeAuthenticator, connect
from xorq.common.utils.env_utils import (
    EnvConfigable,
    env_templates_dir,
    filter_existing_env_vars,
    maybe_substitute_env_var,
)


SnowflakeConfig = EnvConfigable.subclass_from_env_file(
    env_templates_dir.joinpath(".env.snowflake.template")
)
snowflake_config = SnowflakeConfig.from_env()
default_create_object_udfs = bool(snowflake_config.SNOWFLAKE_CREATE_OBJECT_UDFS)
default_database = snowflake_config["SNOWFLAKE_DATABASE"] or "SNOWFLAKE_SAMPLE_DATA"
default_schema = snowflake_config["SNOWFLAKE_SCHEMA"] or "TPCH_SF1"


def make_auth_defaults(authenticator=None):
    match authenticator := str(authenticator).lower():
        case SnowflakeAuthenticator.password:
            return {
                "password": "${SNOWFLAKE_PASSWORD}",
            }
        case SnowflakeAuthenticator.mfa:
            return {
                "password": "${SNOWFLAKE_PASSWORD}",
                "authenticator": authenticator,
            }
        case SnowflakeAuthenticator.keypair:
            return {
                "private_key": "${SNOWFLAKE_PRIVATE_KEY}",
                "private_key_pwd": "${SNOWFLAKE_PRIVATE_KEY_PWD}",
                "authenticator": authenticator,
            }
        case SnowflakeAuthenticator.sso:
            return {
                "authenticator": authenticator,
            }
        case _:
            raise ValueError(f"unsupported Snowflake authenticator: {authenticator!r}")


def make_credential_defaults(authenticator=None):
    dct = {
        "user": "${SNOWFLAKE_USER}",
    } | make_auth_defaults(authenticator)
    return filter_existing_env_vars(dct, snowflake_config)


def make_connection_defaults():
    return {
        "account": snowflake_config["SNOWFLAKE_ACCOUNT"],
        "role": snowflake_config["SNOWFLAKE_ROLE"],
        "warehouse": snowflake_config["SNOWFLAKE_WAREHOUSE"],
    }


def execute_statement(con, statement, do_assert=True):
    fetched = con.raw_sql(statement).fetchall()
    if do_assert:
        assert fetched == [("Statement executed successfully.",)]
    return fetched


def make_connection(
    database=default_database,
    schema=default_schema,
    authenticator=None,
    **kwargs,
):
    con = connect(
        database=f"{database}/{schema}",
        **{
            **make_credential_defaults(authenticator=authenticator),
            **make_connection_defaults(),
            **kwargs,
        },
    )
    return con


def make_raw_connection(database, schema, **kwargs):
    return snowflake.connector.connect(
        database=database,
        schema=schema,
        **{
            **make_credential_defaults(),
            **make_connection_defaults(),
            **kwargs,
        },
    )


def grant_create_database(con, role="public"):
    current_role = con.con.role
    statement = f"""
        USE ROLE accountadmin;
        GRANT CREATE DATABASE ON account TO ROLE {role};
        USE ROLE {current_role}
    """
    execute_statement(con, statement)


def grant_create_schema(con, role="public"):
    current_role = con.con.role
    statement = f"""
        USE ROLE accountadmin;
        GRANT CREATE SCHEMA ON account TO ROLE {role};
        USE ROLE {current_role}
    """
    execute_statement(con, statement)


def get_snowflake_last_modification_time(dt):
    (con, table, database, schema) = (
        dt.source,
        dt.name,
        dt.namespace.catalog,
        dt.namespace.database,
    )
    values = (
        con.table("TABLES", database=(database, "INFORMATION_SCHEMA"))[
            lambda t: t.TABLE_NAME == table
        ][lambda t: t.TABLE_SCHEMA == schema]
        .LAST_ALTERED.execute()
        .values
    )
    if not len(values):
        raise ValueError(
            f"table {table!r} not found in INFORMATION_SCHEMA.TABLES for database {database!r}, schema {schema!r}"
        )
    (value, *rest) = values
    if rest:
        raise ValueError(f"multiple rows returned for table {table!r}")
    return value


def get_grants(con, role="public"):
    data = con.raw_sql(f"SHOW GRANTS TO ROLE {role}").fetchall()
    df = pd.DataFrame(
        data,
        columns=(
            "created_on",
            "privilege",
            "granted_on",
            "name",
            "granted_to",
            "grant_option",
            "granted_by_role_type",
            "granted_by",
        ),
    )
    return df


def get_session_query_df(con):
    stmt = """
    SELECT *
    FROM table(information_schema.query_history_by_session())
    ORDER BY start_time;
    """
    session_query_df = con.raw_sql(stmt).fetch_pandas_all()
    return session_query_df


@frozen
class SnowflakeADBC:
    con = field(validator=instance_of(SnowflakeBackend))

    @property
    def password(self):
        return maybe_substitute_env_var(self.con._profile.kwargs_dict.get("password"))

    @property
    def is_keypair_auth(self):
        return self.con._profile.kwargs_dict.get("authenticator") == "snowflake_jwt"

    @property
    def db_kwargs(self, N=20):
        from xorq.common.utils.snowflake_keypair_utils import SnowflakeKeypair

        if self.is_keypair_auth:
            keypair = SnowflakeKeypair.from_bytes_der(self.con.con._private_key)
            return {
                DatabaseOptions.AUTH_TYPE.value: "auth_jwt",
                DatabaseOptions.JWT_PRIVATE_KEY_VALUE.value: keypair.private_bytes,
                DatabaseOptions.JWT_PRIVATE_KEY_PASSWORD.value: keypair.private_key_pwd,
            }
        else:
            return {}

    @property
    def params(self):
        dct = {
            attr: getattr(self.con.con, attr)
            for attr in ("user", "host", "database", "schema", "warehouse", "role")
        } | {
            # ADBC connection always requires a password, even if using private key
            "password": self.password or "nopassword",
        }
        return dct

    def get_uri(self, **kwargs):
        params = {**self.params, **kwargs}
        uri = f"{params['user']}:{params['password']}@{params['host']}/{params['database']}/{params['schema']}?warehouse={params['warehouse']}&role={params['role']}"
        return uri

    @property
    def uri(self):
        return self.get_uri()

    def get_conn(self, **kwargs):
        return dbapi.connect(
            self.get_uri(**kwargs),
            db_kwargs=self.db_kwargs,
        )

    @property
    def conn(self):
        return self.get_conn()

    def adbc_ingest(
        self,
        table_name,
        record_batch_reader,
        mode="create",
        temporary=False,
        database=None,
        conn_kwargs=(),
        **kwargs,
    ):
        def make_use_stmt(con, catalog, db):
            import sqlglot as sg
            import sqlglot.expressions as sge

            use_stmt = sge.Use(
                kind="SCHEMA",
                this=sg.table(db, catalog=catalog, quoted=con.compiler.quoted),
            ).sql(dialect=con.name)
            return use_stmt

        catalog, db = (
            (self.con.current_catalog, self.con.current_database)
            if database is None
            else database
        )
        # create adbc con pointing to a "known safe" catalog/db
        d = {"database": "SNOWFLAKE_SAMPLE_DATA", "schema": "TPCH_SF1"}
        with self.get_conn(**d | dict(conn_kwargs)) as conn:
            with conn.cursor() as cur:
                cur.execute(make_use_stmt(self.con, catalog, db))
                cur.adbc_ingest(
                    table_name,
                    record_batch_reader,
                    mode=mode,
                    temporary=temporary,
                    **kwargs,
                )
            # must commit!
            conn.commit()



import functools
import operator
import re
from subprocess import (
    PIPE,
    Popen,
)

import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    deep_iterable,
    instance_of,
    or_,
)


@functools.cache
def in_nix_shell():
    return bool(Popened.check_output("echo $IN_NIX_SHELL").strip())


def assert_not_in_nix_shell():
    if in_nix_shell():
        raise ValueError("in nix shell")


try_decode_ascii = toolz.excepts(
    AttributeError, operator.methodcaller("decode", "ascii")
)


try_decode_utf8 = toolz.excepts(
    AttributeError, operator.methodcaller("decode", "utf-8")
)


@frozen(eq=False)
class Popened:
    args = field(
        validator=or_(
            deep_iterable(instance_of(str), instance_of(tuple)), instance_of(str)
        ),
    )
    kwargs_tuple = field(validator=instance_of(tuple), default=())
    deferred = field(validator=instance_of(bool), default=True)

    def __attrs_post_init__(self):
        if isinstance(self.args, str):
            assert self.kwargs.get("shell")
        if not self.deferred:
            self.popen

    @property
    def kwargs(self):
        return dict(self.kwargs_tuple)

    @property
    @functools.cache
    def popen(self):
        popen = non_blocking_subprocess_run(self.args, **self.kwargs)
        return popen

    @property
    @functools.cache
    def stdout_peeker(self):
        from xorq.common.utils.io_utils import Peeker

        return Peeker(self.popen.stdout) if self.popen.stdout else None

    def peek_stdout(self, size):
        return self.stdout_peeker.peek(size)

    @property
    @functools.cache
    def communicated(self):
        # Read already-peeked bytes from the BytesIO buffer (non-blocking)
        peeked = self.stdout_peeker.buf.read() if self.stdout_peeker else b""
        # Always drain both pipes concurrently to avoid deadlock
        (_stdout, _stderr) = self.popen.communicate()
        if peeked:
            _stdout = peeked + (_stdout or b"")
        return (_stdout, _stderr)

    def wait(self):
        self.communicated
        self.popen.wait()

    @property
    def _stdout(self):
        return self.communicated[0]

    @property
    def _stderr(self):
        return self.communicated[1]

    @property
    def stdout(self):
        return try_decode_utf8(self._stdout)

    @property
    def stderr(self):
        return try_decode_utf8(self._stderr)

    @property
    def returncode(self):
        # ensure we have executed
        self.communicated
        return self.popen.returncode

    @classmethod
    def check_output(cls, args, shell=True, **kwargs):
        self = cls(args, tuple(({"shell": shell} | kwargs).items()))
        assert not self.returncode
        return self.stdout


def non_blocking_subprocess_run(args, stdout=PIPE, stderr=PIPE, **kwargs):
    return Popen(args, stdout=stdout, stderr=stderr, **kwargs)


def subprocess_run(args, do_decode=False, **kwargs):
    popened = non_blocking_subprocess_run(args, **kwargs)
    (stdout, stderr) = popened.communicate()
    if do_decode:
        (stdout, stderr) = (try_decode_utf8(el) for el in popened.communicate())
    return (popened.returncode, stdout, stderr)


# https://stackoverflow.com/a/14693789
ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
remove_ansi_escape = functools.partial(ansi_escape.sub, "")



import pathlib

from xorq.ibis_yaml.config import (
    BuildConfig,  # Replace 'your_module' with the appropriate module name
)


def test_default_hash_length():
    config = BuildConfig()
    assert config.hash_length == 12, "Default hash_length should be 12"


def test_default_build_path():
    config = BuildConfig()
    expected_path = pathlib.Path.cwd() / "builds"
    assert config.build_path == expected_path


def test_set_build_path():
    config = BuildConfig()
    new_path = pathlib.Path.cwd() / "custom_builds"
    config.build_path = new_path
    assert config.build_path == new_path



import numpy as np
import pandas as pd
import pytest

import xorq.api as xo


@pytest.fixture
def quotes_df():
    # Quote data (more frequent updates)
    quote_dates = pd.date_range(start="2024-01-01", end="2024-01-31", freq="30min")
    quotes = pd.DataFrame(
        {
            "timestamp": quote_dates,
            "symbol": np.random.choice(["AAPL", "GOOGL", "MSFT"], len(quote_dates)),
            "bid": np.random.uniform(99, 198, len(quote_dates)),
            "ask": np.random.uniform(101, 202, len(quote_dates)),
        }
    )

    return quotes


@pytest.fixture
def trades_df():
    # Create sample trading data
    np.random.seed(42)
    dates = pd.date_range(start="2024-01-01", end="2024-01-31", freq="H")

    # Trades data
    trades = pd.DataFrame(
        {
            "timestamp": dates,
            "symbol": np.random.choice(["AAPL", "GOOGL", "MSFT"], len(dates)),
            "price": np.random.uniform(100, 200, len(dates)),
            "volume": np.random.randint(100, 1000, len(dates)),
        }
    )

    return trades


@pytest.fixture
def ddb_con(quotes_df):
    """Create DuckDB connection with Ibis"""
    duckdb_connection = xo.duckdb.connect()
    duckdb_connection.create_table(
        "quotes",
        quotes_df,
    )
    return duckdb_connection


@pytest.fixture
def con(trades_df):
    """Create DataFusion connection with xorq"""
    con = xo.datafusion.connect()
    con.create_table("trades", trades_df, temp=False)
    return con


@pytest.fixture(scope="session")
def ls_con(parquet_dir):
    conn = xo.connect()

    for name in ("astronauts", "functional_alltypes", "awards_players", "batting"):
        conn.read_parquet(parquet_dir / f"{name}.parquet", name)

    return conn


@pytest.fixture(scope="session")
def alltypes(ls_con):
    return ls_con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def awards_players(ls_con):
    return ls_con.table("awards_players")


@pytest.fixture(scope="session")
def batting(ls_con):
    return ls_con.table("batting")



import operator

import pandas as pd
import pytest
import toolz

import xorq.api as xo
from xorq.caching import ParquetCache
from xorq.expr.udf import make_pandas_udf
from xorq.ibis_yaml.compiler import YamlExpressionTranslator
from xorq.tests.util import assert_frame_equal


field_name = "price_per_carat"
schema = xo.schema({"sum_price": "float64", "sum_carat": "float64"})
return_type = xo.dtype("float64")


@pytest.fixture(scope="session")
def con():
    return xo.connect()


@pytest.fixture(scope="session")
def diamonds(con, parquet_dir):
    return xo.deferred_read_parquet(parquet_dir / "diamonds.parquet", con=con)


@pytest.fixture(scope="session")
def baseline(diamonds):
    expr = diamonds.pipe(do_agg).mutate(my_udf_on_expr).order_by("cut")
    return expr.execute()


@make_pandas_udf(
    schema=schema,
    return_type=return_type,
    name="my_udf",
)
def my_udf(df):
    return df["sum_carat"].div(df["sum_price"])


def do_agg(expr):
    return (
        expr.group_by("cut")
        .agg(
            xo._.price.sum().name("sum_price"),
            xo._.carat.sum().name("sum_carat"),
            xo._.color.nunique().name("nunique_color"),
        )
        .cast({"sum_price": "float64"})
    )


my_udf_on_expr = toolz.compose(
    operator.methodcaller("name", field_name), my_udf.on_expr
)


def test_flight_expr(con, diamonds, baseline):
    unbound_expr = (
        xo.table(diamonds.schema()).pipe(do_agg).mutate(my_udf_on_expr).order_by("cut")
    )
    expr = xo.expr.relations.flight_expr(
        diamonds,
        unbound_expr,
        inner_name="flight-expr",
        name="remote-expr",
        con=con,
    )
    yaml_dict = YamlExpressionTranslator.to_yaml(expr)

    diamonds_con = diamonds._find_backend()

    profiles = {
        con._profile.hash_name: con,
        diamonds_con._profile.hash_name: diamonds_con,
    }
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles)

    df = expr.execute()
    roundtrip_df = roundtrip_expr.execute()
    pd.testing.assert_frame_equal(
        df,
        roundtrip_df,
        check_exact=False,
    )


def test_flight_udxf(con, diamonds, baseline):
    input_expr = diamonds.pipe(do_agg)
    process_df = operator.methodcaller("assign", **{field_name: my_udf.fn})
    maybe_schema_in = input_expr.schema()
    maybe_schema_out = xo.schema(input_expr.schema() | {field_name: return_type})
    expr = xo.expr.relations.flight_udxf(
        input_expr,
        process_df=process_df,
        maybe_schema_in=maybe_schema_in,
        maybe_schema_out=maybe_schema_out,
        con=con,
        make_udxf_kwargs={"name": my_udf.__name__},
    ).order_by("cut")

    yaml_dict = YamlExpressionTranslator.to_yaml(expr)

    diamonds_con = diamonds._find_backend()

    profiles = {
        con._profile.hash_name: con,
        diamonds_con._profile.hash_name: diamonds_con,
    }
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles)

    df = expr.execute()

    roundtrip_df = roundtrip_expr.execute()

    pd.testing.assert_frame_equal(
        df,
        roundtrip_df,
        check_exact=False,
    )


def test_flight_udxf_cached(con, diamonds, baseline):
    from xorq.common.utils.graph_utils import find_all_sources

    input_expr = diamonds.pipe(do_agg)
    process_df = operator.methodcaller("assign", **{field_name: my_udf.fn})
    maybe_schema_in = input_expr.schema()
    maybe_schema_out = xo.schema(input_expr.schema() | {field_name: return_type})

    udxf = xo.expr.relations.flight_udxf(
        process_df=process_df,
        maybe_schema_in=maybe_schema_in,
        maybe_schema_out=maybe_schema_out,
        con=con,
        make_udxf_kwargs={"name": my_udf.__name__},
    )

    raw_expr = input_expr.pipe(udxf)

    ddb_con = xo.duckdb.connect()
    expr = (
        raw_expr.filter(xo._.cut.notnull())
        .cache(cache=ParquetCache.from_kwargs(source=ddb_con))
        .filter(~xo._.cut.contains("ERROR"))
        .order_by("cut")
    )

    yaml_dict = YamlExpressionTranslator.to_yaml(expr)

    profiles = {con._profile.hash_name: con for con in find_all_sources(expr)}
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles)

    expected = expr.execute()

    actual = roundtrip_expr.execute()

    assert_frame_equal(
        expected,
        actual,
        check_exact=False,
    )



import dask
import pandas as pd

import xorq.api as xo
from xorq.backends.conftest import KEY_PREFIX, get_cache_uncached
from xorq.caching import SourceCache, SourceSnapshotCache
from xorq.expr.relations import into_backend
from xorq.vendor import ibis


def test_pandas_snapshot(xo_con, alltypes_df):
    group_by = "year"
    name = ibis.util.gen_name("tmp_table")

    # create a temp table we can mutate
    pd_con = xo.pandas.connect()
    table = pd_con.create_table(name, alltypes_df)

    cached_expr = (
        table.group_by(group_by)
        .agg({f"count_{col}": table[col].count() for col in table.columns})
        .pipe(into_backend, xo_con)
        .cache(cache=SourceSnapshotCache.from_kwargs(source=xo_con))
    )
    (cache, uncached) = get_cache_uncached(cached_expr)

    # test preconditions
    assert not cache.exists(uncached)

    # test cache creation
    executed0 = cached_expr.execute()

    with cache.strategy.normalization_context(uncached):
        normalized0 = dask.base.normalize_token(uncached)
    assert cache.exists(uncached)

    # test cache use
    executed1 = cached_expr.execute()
    assert executed0.equals(executed1)

    # test NO cache invalidation
    pd_con.reconnect()
    table2 = pd_con.create_table(name, pd.concat((alltypes_df, alltypes_df)))

    cached_expr = (
        table2.group_by(group_by)
        .agg({f"count_{col}": table2[col].count() for col in table2.columns})
        .pipe(into_backend, xo_con)
        .cache(cache)
    )
    (cache, uncached) = get_cache_uncached(cached_expr)
    with cache.strategy.normalization_context(uncached):
        normalized1 = dask.base.normalize_token(uncached)

    # everything else is stable, despite the different data
    assert normalized0[1][1] == normalized1[1][1]
    assert cache.exists(uncached)
    assert cache.calc_key(uncached).count(KEY_PREFIX) == 1
    executed2 = cached_expr.ls.uncached.execute()
    assert not executed0.equals(executed2)


def test_caching_pandas(csv_dir):
    diamonds_path = csv_dir / "diamonds.csv"
    pandas_con = xo.pandas.connect()
    cache = SourceCache.from_kwargs(source=pandas_con)
    t = pandas_con.read_csv(diamonds_path).cache(cache=cache)
    assert t.execute() is not None



"""DuckDB backend."""

from __future__ import annotations

import ast
import contextlib
import urllib
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import duckdb
import pyarrow as pa
import pyarrow_hotfix  # noqa: F401
import sqlglot as sg
import sqlglot.expressions as sge
from packaging.version import parse as vparse

import xorq.common.exceptions as exc
import xorq.vendor.ibis.backends.sql.compilers as sc
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends import CanCreateDatabase, CanCreateSchema, UrlFromPath
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.backends.sql.compilers.base import (
    STAR,
    AlterTable,
    C,
    RenameTable,
)
from xorq.vendor.ibis.common.dispatch import lazy_singledispatch
from xorq.vendor.ibis.expr.operations.udf import InputType


if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, MutableMapping, Sequence

    import pandas as pd
    import polars as pl
    import torch
    from fsspec import AbstractFileSystem

    from xorq.vendor.ibis.expr.schema import SchemaLike

try:
    from duckdb import func as _duckdb_func
except ImportError:
    from duckdb import functional as _duckdb_func


try:
    from duckdb import func as _duckdb_func
except ImportError:
    from duckdb import functional as _duckdb_func


_UDF_INPUT_TYPE_MAPPING = {
    InputType.PYARROW: _duckdb_func.ARROW,
    InputType.PYTHON: _duckdb_func.NATIVE,
}


class _Settings:
    def __init__(self, con: duckdb.DuckDBPyConnection) -> None:
        self.con = con

    def __getitem__(self, key: str) -> Any:
        maybe_value = self.con.execute(
            "select value from duckdb_settings() where name = $1", [key]
        ).fetchone()
        if maybe_value is not None:
            return maybe_value[0]
        raise KeyError(key)

    def __setitem__(self, key, value):
        self.con.execute(f"SET {key} = {str(value)!r}")

    def __repr__(self):
        return repr(self.con.sql("from duckdb_settings()"))


class Backend(SQLBackend, CanCreateDatabase, CanCreateSchema, UrlFromPath):
    name = "duckdb"
    compiler = sc.duckdb.compiler
    supports_temporary_tables = True

    @property
    def settings(self) -> _Settings:
        return _Settings(self.con)

    @property
    def current_catalog(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.current_database())) as cur:
            [(db,)] = cur.fetchall()
        return db

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.current_schema())) as cur:
            [(db,)] = cur.fetchall()
        return db

    # TODO(kszucs): should be moved to the base SQLGLot backend
    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.name)
        return self.con.execute(query, **kwargs)

    def create_table(
        self,
        name: str,
        /,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in DuckDB.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.

            For multi-level table hierarchies, you can pass in a dotted string
            path like `"catalog.database"` or a tuple of strings like
            `("catalog", "database")`.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        table_loc = self._to_sqlglot_table(database)

        if getattr(table_loc, "catalog", False) and temp:
            raise exc.UnsupportedArgumentError(
                "DuckDB can only create temporary tables in the `temp` catalog. "
                "Don't specify a catalog to enable temp table creation."
            )

        catalog = table_loc.catalog or self.current_catalog
        database = table_loc.db or self.current_database

        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        quoted = self.compiler.quoted
        dialect = self.dialect

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())
            catalog = "temp"
            database = "main"

        in_memory = False
        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
                in_memory = True
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if schema is None:
            schema = table.schema()
        else:
            schema = ibis.schema(schema)

        if overwrite:
            temp_name = util.gen_name("duckdb_table")
        else:
            temp_name = name

        initial_table = sg.table(temp_name, catalog=catalog, db=database, quoted=quoted)
        target = sge.Schema(
            this=initial_table,
            expressions=(schema or table.schema()).to_sqlglot(self.dialect),
        )

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        # This is the same table as initial_table unless overwrite == True
        final_table = sg.table(name, catalog=catalog, db=database, quoted=quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                columns = [
                    sge.to_identifier(col, quoted=quoted) for col in table.columns
                ]
                insert_stmt = sge.insert(
                    query, into=initial_table, columns=columns
                ).sql(dialect)
                cur.execute(insert_stmt).fetchall()

                if in_memory:
                    cur.execute(
                        sge.Drop(kind="VIEW", this=table.get_name(), exists=True).sql(
                            dialect
                        )
                    )

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=final_table, exists=True).sql(dialect)
                )
                # TODO: This branching should be removed once DuckDB >=0.9.3 is
                # our lower bound (there's an upstream bug in 0.9.2 that
                # disallows renaming temp tables)
                # We should (pending that release) be able to remove the if temp
                # branch entirely.
                if temp:
                    cur.execute(
                        sge.Create(
                            kind="TABLE",
                            this=final_table,
                            expression=sg.select(STAR).from_(initial_table),
                            properties=sge.Properties(expressions=properties),
                        ).sql(dialect)
                    )
                    cur.execute(
                        sge.Drop(kind="TABLE", this=initial_table, exists=True).sql(
                            dialect
                        )
                    )
                else:
                    cur.execute(
                        AlterTable(
                            this=initial_table,
                            actions=[RenameTable(this=final_table)],
                        ).sql(dialect)
                    )

        return self.table(name, database=(catalog, database))

    def table(self, name: str, /, *, database: str | None = None) -> ir.Table:
        table_loc = self._to_sqlglot_table(database)

        # TODO: set these to better defaults
        catalog = table_loc.catalog or None
        database = table_loc.db or None

        table_schema = self.get_schema(name, catalog=catalog, database=database)
        # load geospatial only if geo columns
        if table_schema.geospatial:
            self.load_extension("spatial")
        return ops.DatabaseTable(
            name,
            schema=table_schema,
            source=self,
            namespace=ops.Namespace(catalog=catalog, database=database),
        ).to_expr()

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        """Compute the schema of a `table`.

        Parameters
        ----------
        table_name
            May **not** be fully qualified. Use `database` if you want to
            qualify the identifier.
        catalog
            Catalog name
        database
            Database name

        Returns
        -------
        sch.Schema
            Ibis schema
        """
        query = sge.Describe(
            this=sg.table(
                table_name, db=database, catalog=catalog, quoted=self.compiler.quoted
            )
        ).sql(self.dialect)

        try:
            result = self.con.sql(query)
        except duckdb.CatalogException:
            raise exc.XorqError(f"Table not found: {table_name!r}")
        else:
            meta = result.fetch_arrow_table()

        names = meta["column_name"].to_pylist()
        types = meta["column_type"].to_pylist()
        nullables = meta["null"].to_pylist()

        type_mapper = self.compiler.type_mapper
        return sch.Schema(
            {
                name: type_mapper.from_string(typ, nullable=null == "YES")
                for name, typ, null in zip(names, types, nullables)
            }
        )

    @contextlib.contextmanager
    def _safe_raw_sql(self, *args, **kwargs):
        yield self.raw_sql(*args, **kwargs)

    def list_catalogs(self, *, like: str | None = None) -> list[str]:
        col = "catalog_name"
        query = sg.select(sge.Distinct(expressions=[sg.column(col)])).from_(
            sg.table("schemata", db="information_schema")
        )
        with self._safe_raw_sql(query) as cur:
            result = cur.fetch_arrow_table()
        dbs = result[col]
        return self._filter_with_like(dbs.to_pylist(), like)

    def list_databases(
        self, *, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        col = "schema_name"
        query = sg.select(sge.Distinct(expressions=[sg.column(col)])).from_(
            sg.table("schemata", db="information_schema")
        )

        if catalog is not None:
            query = query.where(sg.column("catalog_name").eq(sge.convert(catalog)))

        with self._safe_raw_sql(query) as cur:
            out = cur.fetch_arrow_table()
        return self._filter_with_like(out[col].to_pylist(), like=like)

    @staticmethod
    def _convert_kwargs(kwargs: MutableMapping) -> None:
        read_only = str(kwargs.pop("read_only", "False")).capitalize()
        try:
            kwargs["read_only"] = ast.literal_eval(read_only)
        except ValueError as e:
            raise ValueError(
                f"invalid value passed to ast.literal_eval: {read_only!r}"
            ) from e

    @property
    def version(self) -> str:
        # TODO: there is a `PRAGMA version` we could use instead
        import importlib.metadata

        return importlib.metadata.version("duckdb")

    def do_connect(
        self,
        database: str | Path = ":memory:",
        read_only: bool = False,
        extensions: Sequence[str] | None = None,
        **config: Any,
    ) -> None:
        """Create an Ibis client connected to a DuckDB database.

        ::: {.callout-note title="Changed in version 10.0.0"}
        Before, we had special handling if the user passed the `temp_directory`
        parameter, setting a custom default, and creating intermediate
        directories if necessary. Now, we do nothing, and just pass the value
        directly to DuckDB. You may need to add
        `Path(your_temp_dir).mkdir(exists_ok=True, parents=True)`
        to your code to maintain the old behavior.
        :::

        Parameters
        ----------
        database
            Path to a duckdb database.
        read_only
            Whether the database is read-only.
        extensions
            A list of duckdb extensions to install/load upon connection.
        config
            DuckDB configuration parameters. See the [DuckDB configuration
            documentation](https://duckdb.org/docs/sql/configuration) for
            possible configuration values.

        Examples
        --------
        >>> import ibis
        >>> ibis.duckdb.connect(threads=4, memory_limit="1GB")  # doctest: +ELLIPSIS
        <ibis.backends.duckdb.Backend object at 0x...>
        """
        if not isinstance(database, Path) and not database.startswith(
            ("md:", "motherduck:", ":memory:")
        ):
            database = Path(database).absolute()
        self.con = duckdb.connect(str(database), config=config, read_only=read_only)
        self._post_connect(extensions)

    @util.experimental
    @classmethod
    def from_connection(
        cls,
        con: duckdb.DuckDBPyConnection,
        /,
        *,
        extensions: Sequence[str] | None = None,
    ) -> Backend:
        """Create an Ibis client from an existing connection to a DuckDB database.

        Parameters
        ----------
        con
            An existing connection to a DuckDB database.
        extensions
            A list of duckdb extensions to install/load upon connection.
        """
        new_backend = cls(extensions=extensions)
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect(extensions)
        return new_backend

    def _post_connect(self, extensions: Sequence[str] | None = None) -> None:
        # Load any pre-specified extensions
        if extensions is not None:
            self._load_extensions(extensions)

        # Default timezone, can't be set with `config`
        self.settings["timezone"] = "UTC"

        # setting this to false disables magic variables-as-tables discovery,
        # hopefully eliminating large classes of bugs
        if vparse(self.version) > vparse("1"):
            self.settings["python_enable_replacements"] = False

        self._record_batch_readers_consumed = {}

    def _load_extensions(
        self, extensions: list[str], force_install: bool = False
    ) -> None:
        f = self.compiler.f
        query = (
            sg.select(
                f.anon.unnest(
                    f.list_intersect(
                        f.list_append(C.aliases, C.extension_name),
                        f.list_value(*extensions),
                    )
                ),
                C.installed,
                C.loaded,
            )
            .from_(f.duckdb_extensions())
            .where(sg.not_(C.installed & C.loaded))
        )
        with self._safe_raw_sql(query) as cur:
            if not (not_installed_or_loaded := cur.fetchall()):
                return

            commands = [
                "FORCE " * force_install + f"INSTALL '{extension}'"
                for extension, installed, _ in not_installed_or_loaded
                if not installed
            ]
            commands.extend(
                f"LOAD '{extension}'"
                for extension, _, loaded in not_installed_or_loaded
                if not loaded
            )
            command = ";".join(commands)
            cur.execute(command)

    def load_extension(self, extension: str, force_install: bool = False) -> None:
        """Install and load a duckdb extension by name or path.

        Parameters
        ----------
        extension
            The extension name or path.
        force_install
            Force reinstallation of the extension.

        """
        self._load_extensions([extension], force_install=force_install)

    def create_database(
        self, name: str, /, *, catalog: str | None = None, force: bool = False
    ) -> None:
        if catalog is not None:
            raise exc.UnsupportedOperationError(
                "DuckDB cannot create a database in another catalog."
            )

        name = sg.table(name, catalog=catalog, quoted=self.compiler.quoted)
        with self._safe_raw_sql(sge.Create(this=name, kind="SCHEMA", replace=force)):
            pass

    def drop_database(
        self, name: str, /, *, catalog: str | None = None, force: bool = False
    ) -> None:
        if catalog is not None and catalog != self.current_catalog:
            raise exc.UnsupportedOperationError(
                "DuckDB cannot drop a database in another catalog."
            )

        name = sg.table(name, catalog=catalog, quoted=self.compiler.quoted)
        with self._safe_raw_sql(sge.Drop(this=name, kind="SCHEMA", replace=force)):
            pass

    @util.experimental
    def read_json(
        self,
        paths: str | list[str] | tuple[str],
        /,
        *,
        table_name: str | None = None,
        columns: Mapping[str, str] | None = None,
        **kwargs,
    ) -> ir.Table:
        """Read newline-delimited JSON into an ibis table.

        ::: {.callout-note}
        ## This feature requires duckdb>=0.7.0
        :::

        Parameters
        ----------
        paths
            File or list of files
        table_name
            Optional table name
        columns
            Optional mapping from string column name to duckdb type string.
        **kwargs
            Additional keyword arguments passed to DuckDB's `read_json_auto` function.

            See https://duckdb.org/docs/data/json/overview.html#json-loading
            for parameters and more information about reading JSON.

        Returns
        -------
        Table
            An ibis table expression

        """
        if not table_name:
            table_name = util.gen_name("read_json")

        options = [
            sg.to_identifier(key).eq(sge.convert(val)) for key, val in kwargs.items()
        ]

        if columns:
            options.append(
                sg.to_identifier("columns").eq(
                    sge.Struct.from_arg_list(
                        [
                            sge.PropertyEQ(
                                this=sg.to_identifier(key),
                                expression=sge.convert(value),
                            )
                            for key, value in columns.items()
                        ]
                    )
                )
            )

        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(
                self.compiler.f.read_json_auto(
                    util.normalize_filenames(paths), *options
                )
            ),
        )

        return self.table(table_name)

    def read_csv(
        self,
        paths: str | list[str] | tuple[str],
        /,
        *,
        table_name: str | None = None,
        columns: Mapping[str, str | dt.DataType] | None = None,
        types: Mapping[str, str | dt.DataType] | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a CSV file as a table in the current database.

        Parameters
        ----------
        paths
            The data source(s). May be a path to a file or directory of CSV
            files, or an iterable of CSV files.
        table_name
            An optional name to use for the created table. This defaults to a
            sequentially generated name.
        columns
            An optional mapping of **all** column names to their types.
        types
            An optional mapping of a **subset** of column names to their types.
        **kwargs
            Additional keyword arguments passed to DuckDB loading function. See
            https://duckdb.org/docs/data/csv for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        Examples
        --------
        Generate some data

        >>> import tempfile
        >>> data = b'''
        ... lat,lon,geom
        ... 1.0,2.0,POINT (1 2)
        ... 2.0,3.0,POINT (2 3)
        ... '''
        >>> with tempfile.NamedTemporaryFile(delete=False) as f:
        ...     nbytes = f.write(data)

        Import Ibis

        >>> import ibis
        >>> from ibis import _
        >>> ibis.options.interactive = True
        >>> con = ibis.duckdb.connect()

        Read the raw CSV file

        >>> t = con.read_csv(f.name)
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━┓
        ┃ lat     ┃ lon     ┃ geom        ┃
        ┡━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━┩
        │ float64 │ float64 │ string      │
        ├─────────┼─────────┼─────────────┤
        │     1.0 │     2.0 │ POINT (1 2) │
        │     2.0 │     3.0 │ POINT (2 3) │
        └─────────┴─────────┴─────────────┘

        Load the `spatial` extension and read the CSV file again, using
        specific column types

        >>> con.load_extension("spatial")
        >>> t = con.read_csv(f.name, types={"geom": "geometry"})
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ lat     ┃ lon     ┃ geom                 ┃
        ┡━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
        │ float64 │ float64 │ geospatial:geometry  │
        ├─────────┼─────────┼──────────────────────┤
        │     1.0 │     2.0 │ <POINT (1 2)>        │
        │     2.0 │     3.0 │ <POINT (2 3)>        │
        └─────────┴─────────┴──────────────────────┘
        """
        paths = util.normalize_filenames(paths)

        if not table_name:
            table_name = util.gen_name("read_csv")

        # auto_detect and columns collide, so we set auto_detect=True
        # unless COLUMNS has been specified
        if any(source.startswith(("http://", "https://", "s3://")) for source in paths):
            self._load_extensions(["httpfs"])

        kwargs.setdefault("header", True)
        kwargs["auto_detect"] = kwargs.pop("auto_detect", columns is None)
        # TODO: clean this up
        # We want to _usually_ quote arguments but if we quote `columns` it messes
        # up DuckDB's struct parsing.
        options = [C[key].eq(sge.convert(val)) for key, val in kwargs.items()]

        def make_struct_argument(obj: Mapping[str, str | dt.DataType]) -> sge.Struct:
            expressions = []
            geospatial = False
            type_mapper = self.compiler.type_mapper

            for name, typ in obj.items():
                typ = dt.dtype(typ)
                geospatial |= typ.is_geospatial()
                sgtype = type_mapper.from_ibis(typ)
                prop = sge.PropertyEQ(
                    this=sge.to_identifier(name), expression=sge.convert(sgtype)
                )
                expressions.append(prop)

            if geospatial:
                self._load_extensions(["spatial"])
            return sge.Struct(expressions=expressions)

        if columns is not None:
            options.append(C.columns.eq(make_struct_argument(columns)))

        if types is not None:
            options.append(C.types.eq(make_struct_argument(types)))

        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(self.compiler.f.read_csv(paths, *options)),
        )

        return self.table(table_name)

    def read_geo(
        self, path: str, /, *, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a geospatial data file as a table in the current database.

        Parameters
        ----------
        path
            The data source(s). Path to a file of geospatial files supported by duckdb.
            See https://duckdb.org/docs/extensions/spatial.html#st_read---read-spatial-data-from-files
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        kwargs
            Additional keyword arguments passed to DuckDB loading function.
            See https://duckdb.org/docs/extensions/spatial.html#st_read---read-spatial-data-from-files
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """

        if not table_name:
            table_name = util.gen_name("read_geo")

        # load geospatial extension
        self.load_extension("spatial")

        path = util.normalize_filename(path)
        if path.startswith(("http://", "https://", "s3://")):
            self._load_extensions(["httpfs"])

        source_expr = sg.select(STAR).from_(
            self.compiler.f.st_read(
                path,
                *(sg.to_identifier(key).eq(val) for key, val in kwargs.items()),
            )
        )

        view = sge.Create(
            kind="VIEW",
            this=sg.table(table_name, quoted=self.compiler.quoted),
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
            expression=source_expr,
        )
        with self._safe_raw_sql(view):
            pass
        return self.table(table_name)

    def read_parquet(
        self,
        paths: str | Path | Iterable[str | Path],
        /,
        *,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        paths
            The data source(s). May be a path to a file, an iterable of files,
            or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to DuckDB loading function.
            See https://duckdb.org/docs/data/parquet for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        paths = util.normalize_filenames(paths)

        table_name = table_name or util.gen_name("read_parquet")

        # Default to using the native duckdb parquet reader
        # If that fails because of auth issues, fall back to ingesting via
        # pyarrow dataset
        try:
            self._read_parquet_duckdb_native(paths, table_name, **kwargs)
        except duckdb.IOException:
            self._read_parquet_pyarrow_dataset(paths, table_name, **kwargs)

        return self.table(table_name)

    def _read_parquet_duckdb_native(
        self, source_list: str | Iterable[str], table_name: str, **kwargs: Any
    ) -> None:
        if any(
            source.startswith(("http://", "https://", "s3://"))
            for source in source_list
        ):
            self._load_extensions(["httpfs"])

        options = [
            sg.to_identifier(key).eq(sge.convert(val)) for key, val in kwargs.items()
        ]
        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(self.compiler.f.read_parquet(source_list, *options)),
        )

    def _read_parquet_pyarrow_dataset(
        self, source_list: str | Iterable[str], table_name: str, **kwargs: Any
    ) -> None:
        import pyarrow.dataset as ds

        dataset = ds.dataset(list(map(ds.dataset, source_list)), **kwargs)
        self._load_extensions(["httpfs"])
        # We don't create a view since DuckDB special cases Arrow Datasets
        # so if we also create a view we end up with both a "lazy table"
        # and a view with the same name
        self.con.register(table_name, dataset)
        # DuckDB normally auto-detects Arrow Datasets that are defined
        # in local variables but the `dataset` variable won't be local
        # by the time we execute against this so we register it
        # explicitly.

    @util.deprecated(
        instead="Pass in-memory data to `memtable` instead.",
        as_of="9.1",
        removed_in="10.0",
    )
    def read_in_memory(
        self,
        source: pd.DataFrame
        | pa.Table
        | pa.RecordBatchReader
        | pl.DataFrame
        | pl.LazyFrame,
        table_name: str | None = None,
    ) -> ir.Table:
        """Register an in-memory table object in the current database.

        Supported objects include pandas DataFrame, a Polars
        DataFrame/LazyFrame, or a PyArrow Table or RecordBatchReader.

        Parameters
        ----------
        source
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_in_memory")
        _read_in_memory(source, table_name, self)
        return self.table(table_name)

    def read_delta(
        self, path: str | Path, /, *, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Delta Lake table as a table in the current database.

        Parameters
        ----------
        path
            The data source. Must be a directory containing a Delta Lake table.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        kwargs
            Additional keyword arguments passed to deltalake.DeltaTable.

        Returns
        -------
        ir.Table
            The just-registered table.
        """
        (path,) = util.normalize_filenames(path)

        extensions = ["delta"]
        if path.startswith(("http://", "https://", "s3://")):
            extensions.append("httpfs")

        table_name = table_name or util.gen_name("read_delta")

        options = [
            sg.to_identifier(key).eq(sge.convert(val)) for key, val in kwargs.items()
        ]

        self._load_extensions(extensions)

        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(self.compiler.f.delta_scan(path, *options)),
        )
        return self.table(table_name)

    def list_tables(
        self, *, like: str | None = None, database: tuple[str, str] | str | None = None
    ) -> list[str]:
        table_loc = self._to_sqlglot_table(database)

        catalog = table_loc.catalog or self.current_catalog
        database = table_loc.db or self.current_database

        col = "table_name"
        sql = (
            sg.select(col)
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
            .where(
                C.table_catalog.isin(sge.convert(catalog), sge.convert("temp")),
                C.table_schema.eq(sge.convert(database)),
            )
            .sql(self.dialect)
        )
        out = self.con.execute(sql).fetch_arrow_table()

        return self._filter_with_like(out[col].to_pylist(), like)

    def read_postgres(
        self, uri: str, /, *, table_name: str | None = None, database: str = "public"
    ) -> ir.Table:
        """Register a table from a postgres instance into a DuckDB table.

        ::: {.callout-note}
        ## Ibis does not use the word `schema` to refer to database hierarchy.

        A collection of `table` is referred to as a `database`.
        A collection of `database` is referred to as a `catalog`.

        These terms are mapped onto the corresponding features in each
        backend (where available), regardless of whether the backend itself
        uses the same terminology.
        :::

        Parameters
        ----------
        uri
            A postgres URI of the form `postgres://user:password@host:port`
        table_name
            The table to read
        database
            PostgreSQL database (schema) where `table_name` resides

        Returns
        -------
        ir.Table
            The just-registered table.

        """
        if table_name is None:
            raise ValueError(
                "`table_name` is required when registering a postgres table"
            )
        self._load_extensions(["postgres_scanner"])

        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(
                self.compiler.f.postgres_scan_pushdown(uri, database, table_name)
            ),
        )

        return self.table(table_name)

    def read_mysql(
        self,
        uri: str,
        /,
        *,
        catalog: str,
        table_name: str | None = None,
    ) -> ir.Table:
        """Register a table from a MySQL instance into a DuckDB table.

        Parameters
        ----------
        uri
            A mysql URI of the form `mysql://user:password@host:port/database`
        catalog
            User-defined alias given to the MySQL database that is being attached
            to DuckDB
        table_name
            The table to read

        Returns
        -------
        ir.Table
            The just-registered table.
        """

        parsed = urllib.parse.urlparse(uri)

        if table_name is None:
            raise ValueError("`table_name` is required when registering a mysql table")

        self._load_extensions(["mysql"])

        database = parsed.path.strip("/")

        query_con = f"""ATTACH 'host={parsed.hostname} user={parsed.username} password={parsed.password} port={parsed.port} database={database}' AS {catalog} (TYPE mysql)"""

        with self._safe_raw_sql(query_con):
            pass

        return self.table(table_name, database=(catalog, database))

    def read_sqlite(
        self, path: str | Path, /, *, table_name: str | None = None
    ) -> ir.Table:
        """Register a table from a SQLite database into a DuckDB table.

        Parameters
        ----------
        path
            The path to the SQLite database
        table_name
            The table to read

        Returns
        -------
        ir.Table
            The just-registered table.

        Examples
        --------
        >>> import ibis
        >>> import sqlite3
        >>> ibis.options.interactive = True
        >>> con = sqlite3.connect("/tmp/sqlite.db")
        >>> with con:
        ...     con.execute("DROP TABLE IF EXISTS t")  # doctest: +ELLIPSIS
        ...     con.execute("CREATE TABLE t (a INT, b TEXT)")  # doctest: +ELLIPSIS
        ...     con.execute(
        ...         "INSERT INTO t VALUES (1, 'a'), (2, 'b'), (3, 'c')"
        ...     )  # doctest: +ELLIPSIS
        <...>
        >>> con.close()
        >>> con = ibis.connect("duckdb://")
        >>> t = con.read_sqlite("/tmp/sqlite.db", table_name="t")
        >>> t
        ┏━━━━━━━┳━━━━━━━━┓
        ┃ a     ┃ b      ┃
        ┡━━━━━━━╇━━━━━━━━┩
        │ int64 │ string │
        ├───────┼────────┤
        │     1 │ a      │
        │     2 │ b      │
        │     3 │ c      │
        └───────┴────────┘

        """

        if table_name is None:
            raise ValueError("`table_name` is required when registering a sqlite table")
        self._load_extensions(["sqlite"])

        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(
                self.compiler.f.sqlite_scan(
                    sg.to_identifier(str(path), quoted=True), table_name
                )
            ),
        )

        return self.table(table_name)

    @util.experimental
    def read_xlsx(
        self,
        path: str | Path,
        /,
        *,
        sheet: str | None = None,
        range: str | None = None,
        **kwargs,
    ) -> ir.Table:
        """Read an Excel file into a DuckDB table. This requires duckdb>=1.2.0.

        Parameters
        ----------
        path
            The path to the Excel file.
        sheet
            The name of the sheet to read, eg 'Sheet3'.
        range
            The range of cells to read, eg 'A5:Z'.
        kwargs
            Additional args passed to the backend's read function.

        Returns
        -------
        ir.Table
            The just-registered table.

        See Also
        --------
        [DuckDB's `excel` extension docs for reading](https://duckdb.org/docs/stable/extensions/excel.html#reading-xlsx-files)

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> t = ibis.memtable({"a": [1, 2, 3], "b": ["a", "b", "c"]})
        >>> con = ibis.duckdb.connect()
        >>> con.to_xlsx(t, "/tmp/test.xlsx", header=True)
        >>> assert os.path.exists("/tmp/test.xlsx")
        >>> t = con.read_xlsx("/tmp/test.xlsx")
        >>> t.columns
        ('a', 'b')
        """
        table_name = util.gen_name("read_xlsx")

        if sheet:
            kwargs["sheet"] = sheet

        if range:
            kwargs["range"] = range

        options = [
            sg.to_identifier(key).eq(sge.convert(val)) for key, val in kwargs.items()
        ]

        self.load_extension("excel")
        self._create_temp_view(
            table_name,
            sg.select(STAR).from_(self.compiler.f.read_xlsx(str(path), *options)),
        )
        return self.table(table_name)

    def to_xlsx(
        self,
        expr: ir.Table,
        /,
        path: str | Path,
        *,
        sheet: str = "Sheet1",
        header: bool = False,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ):
        """Write a table to an Excel file.

        Parameters
        ----------
        expr
            Ibis table expression to write to an excel file.
        path
            Excel output path.
        sheet
            The name of the sheet to write to, eg 'Sheet3'.
        header
            Whether to include the column names as the first row.
        params
            Additional Ibis expression parameters to pass to the backend's
            write function.
        kwargs
            Additional arguments passed to the backend's write function.

        Notes
        -----
        Requires DuckDB >= 1.2.0.

        See Also
        --------
        [DuckDB's `excel` extension docs for writing](https://duckdb.org/docs/stable/extensions/excel.html#writing-xlsx-files)

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> t = ibis.memtable({"a": [1, 2, 3], "b": ["a", "b", "c"]})
        >>> con = ibis.duckdb.connect()
        >>> con.to_xlsx(t, "/tmp/test.xlsx")
        >>> os.path.exists("/tmp/test.xlsx")
        True
        """
        self._run_pre_execute_hooks(expr)
        query = self.compile(expr, params=params)
        kwargs["sheet"] = sheet
        kwargs["header"] = header
        args = ["FORMAT 'xlsx'"]
        args.extend(f"{k.upper()} {v!r}" for k, v in kwargs.items())
        copy_cmd = f"COPY ({query}) TO {str(path)!r} ({', '.join(args)})"
        self.load_extension("excel")
        self.con.execute(copy_cmd).fetchall()

    def attach(
        self, path: str | Path, name: str | None = None, read_only: bool = False
    ) -> None:
        """Attach another DuckDB database to the current DuckDB session.

        Parameters
        ----------
        path
            Path to the database to attach.
        name
            Name to attach the database as. Defaults to the basename of `path`.
        read_only
            Whether to attach the database as read-only.

        """
        code = f"ATTACH '{path}'"

        if name is not None:
            name = sg.to_identifier(name).sql(self.name)
            code += f" AS {name}"

        if read_only:
            code += " (READ_ONLY)"

        self.con.execute(code).fetchall()

    def detach(self, name: str) -> None:
        """Detach a database from the current DuckDB session.

        Parameters
        ----------
        name
            The name of the database to detach.

        """
        name = sg.to_identifier(name).sql(self.name)
        self.con.execute(f"DETACH {name}").fetchall()

    def attach_sqlite(
        self, path: str | Path, overwrite: bool = False, all_varchar: bool = False
    ) -> None:
        """Attach a SQLite database to the current DuckDB session.

        Parameters
        ----------
        path
            The path to the SQLite database.
        overwrite
            Allow overwriting any tables or views that already exist in your current
            session with the contents of the SQLite database.
        all_varchar
            Set all SQLite columns to type `VARCHAR` to avoid type errors on ingestion.

        Examples
        --------
        >>> import ibis
        >>> import sqlite3
        >>> con = sqlite3.connect("/tmp/attach_sqlite.db")
        >>> with con:
        ...     con.execute("DROP TABLE IF EXISTS t")  # doctest: +ELLIPSIS
        ...     con.execute("CREATE TABLE t (a INT, b TEXT)")  # doctest: +ELLIPSIS
        ...     con.execute(
        ...         "INSERT INTO t VALUES (1, 'a'), (2, 'b'), (3, 'c')"
        ...     )  # doctest: +ELLIPSIS
        <...>
        >>> con.close()
        >>> con = ibis.connect("duckdb://")
        >>> con.list_tables()
        []
        >>> con.attach_sqlite("/tmp/attach_sqlite.db")
        >>> con.list_tables()
        ['t']

        """
        self.load_extension("sqlite")
        with self._safe_raw_sql(f"SET GLOBAL sqlite_all_varchar={all_varchar}") as cur:
            cur.execute(
                f"CALL sqlite_attach('{path}', overwrite={overwrite})"
            ).fetchall()

    def register_filesystem(self, filesystem: AbstractFileSystem):
        """Register an `fsspec` filesystem object with DuckDB.

        This allow a user to read from any `fsspec` compatible filesystem using
        `read_csv`, `read_parquet`, `read_json`, etc.


        ::: {.callout-note}
        Creating an `fsspec` filesystem requires that the corresponding
        backend-specific `fsspec` helper library is installed.

        e.g. to connect to Google Cloud Storage, `gcsfs` must be installed.
        :::

        Parameters
        ----------
        filesystem
            The fsspec filesystem object to register with DuckDB.
            See https://duckdb.org/docs/guides/python/filesystems for details.

        Examples
        --------
        >>> import ibis
        >>> import fsspec
        >>> ibis.options.interactive = True
        >>> gcs = fsspec.filesystem("gcs")
        >>> con = ibis.duckdb.connect()
        >>> con.register_filesystem(gcs)
        >>> t = con.read_csv(
        ...     "gcs://ibis-examples/data/band_members.csv.gz",
        ...     table_name="band_members",
        ... )
        >>> t
        ┏━━━━━━━━┳━━━━━━━━━┓
        ┃ name   ┃ band    ┃
        ┡━━━━━━━━╇━━━━━━━━━┩
        │ string │ string  │
        ├────────┼─────────┤
        │ Mick   │ Stones  │
        │ John   │ Beatles │
        │ Paul   │ Beatles │
        └────────┴─────────┘
        """
        self.con.register_filesystem(filesystem)

    def _run_pre_execute_hooks(self, expr: ir.Expr) -> None:
        # Warn for any tables depending on RecordBatchReaders that have already
        # started being consumed.
        for t in expr.op().find(ops.PhysicalTable):
            started = self._record_batch_readers_consumed.get(t.name)
            if started is True:
                warnings.warn(
                    f"Table {t.name!r} is backed by a `pyarrow.RecordBatchReader` "
                    "that has already been partially consumed. This may lead to "
                    "unexpected results. Either recreate the table from a new "
                    "`pyarrow.RecordBatchReader`, or use `Table.cache()`/"
                    "`con.create_table()` to consume and store the results in "
                    "the backend to reuse later."
                )
            elif started is False:
                self._record_batch_readers_consumed[t.name] = True

        if expr.op().find((ops.GeoSpatialUnOp, ops.GeoSpatialBinOp)):
            self.load_extension("spatial")

        super()._run_pre_execute_hooks(expr)

    def _to_duckdb_relation(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ):
        """Preprocess the expr, and return a `duckdb.DuckDBPyRelation` object.

        When retrieving in-memory results, it's faster to use `duckdb_con.sql`
        than `duckdb_con.execute`, as the query planner can take advantage of
        knowing the output type. Since the relation objects aren't compatible
        with the dbapi, we choose to only use them in select internal methods
        where performance might matter, and use the standard
        `duckdb_con.execute` everywhere else.
        """
        self._run_pre_execute_hooks(expr)
        table_expr = expr.as_table()
        sql = self.compile(table_expr, limit=limit, params=params, **kwargs)
        if table_expr.schema().geospatial:
            self._load_extensions(["spatial"])
        return self.con.sql(sql)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        """Return a stream of record batches.

        The returned `RecordBatchReader` contains a cursor with an unbounded lifetime.

        For analytics use cases this is usually nothing to fret about. In some cases you
        may need to explicit release the cursor.

        Parameters
        ----------
        expr
            Ibis expression
        params
            Bound parameters
        limit
            Limit the result to this number of rows
        chunk_size
            The number of rows to fetch per batch
        """
        import pyarrow as pa

        self._run_pre_execute_hooks(expr)
        table = expr.as_table()
        sql = self.compile(table, limit=limit, params=params)

        def batch_producer(cur):
            yield from cur.fetch_record_batch(rows_per_batch=chunk_size)

        result = self.raw_sql(sql)
        return pa.ipc.RecordBatchReader.from_batches(
            expr.as_table().schema().to_pyarrow(), batch_producer(result)
        )

    def to_pyarrow(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        from ibis.backends.duckdb.converter import DuckDBPyArrowData

        table = self._to_duckdb_relation(
            expr, params=params, limit=limit, **kwargs
        ).to_arrow_table()
        return expr.__pyarrow_result__(table, data_mapper=DuckDBPyArrowData)

    def execute(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pd.DataFrame | pd.Series | Any:
        """Execute an expression."""
        import pandas as pd
        import pyarrow.types as pat
        from ibis.backends.duckdb.converter import DuckDBPandasData

        rel = self._to_duckdb_relation(expr, params=params, limit=limit, **kwargs)
        table = rel.to_arrow_table()

        df = pd.DataFrame(
            {
                name: (
                    col.to_pylist()
                    if (
                        pat.is_nested(col.type)
                        or pat.is_dictionary(col.type)
                        or
                        # pyarrow / duckdb type null literals columns as int32?
                        # but calling `to_pylist()` will render it as None
                        col.null_count
                    )
                    else col.to_pandas()
                )
                for name, col in zip(table.column_names, table.columns)
            }
        )
        df = DuckDBPandasData.convert_table(df, expr.as_table().schema())
        return expr.__pandas_result__(df)

    @util.experimental
    def to_torch(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> dict[str, torch.Tensor]:
        """Execute an expression and return results as a dictionary of torch tensors.

        Parameters
        ----------
        expr
            Ibis expression to execute.
        params
            Parameters to substitute into the expression.
        limit
            An integer to effect a specific row limit. A value of `None` means no limit.
        kwargs
            Keyword arguments passed into the backend's `to_torch` implementation.

        Returns
        -------
        dict[str, torch.Tensor]
            A dictionary of torch tensors, keyed by column name.

        """
        return self._to_duckdb_relation(
            expr, params=params, limit=limit, **kwargs
        ).torch()

    @util.experimental
    def to_parquet(
        self,
        expr: ir.Table,
        /,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a parquet file.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to parquet.
        path
            The data source. A string or Path to the parquet file.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            DuckDB Parquet writer arguments. See
            https://duckdb.org/docs/data/parquet/overview.html#writing-to-parquet-files
            for details.

        Examples
        --------
        Write out an expression to a single parquet file.

        >>> import ibis
        >>> penguins = ibis.examples.penguins.fetch()
        >>> con = ibis.get_backend(penguins)
        >>> con.to_parquet(penguins, "/tmp/penguins.parquet")

        Write out an expression to a hive-partitioned parquet file.

        >>> import tempfile
        >>> penguins = ibis.examples.penguins.fetch()
        >>> con = ibis.get_backend(penguins)

        Partition on a single column.

        >>> con.to_parquet(penguins, tempfile.mkdtemp(), partition_by="year")

        Partition on multiple columns.

        >>> con.to_parquet(penguins, tempfile.mkdtemp(), partition_by=("year", "island"))

        """
        self._run_pre_execute_hooks(expr)
        query = self.compile(expr, params=params)
        args = ["FORMAT 'parquet'", *(f"{k.upper()} {v!r}" for k, v in kwargs.items())]
        copy_cmd = f"COPY ({query}) TO {str(path)!r} ({', '.join(args)})"
        with self._safe_raw_sql(copy_cmd):
            pass

    @util.experimental
    def to_csv(
        self,
        expr: ir.Table,
        /,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        header: bool = True,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a CSV file.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to CSV.
        path
            The data source. A string or Path to the CSV file.
        params
            Mapping of scalar parameter expressions to value.
        header
            Whether to write the column names as the first line of the CSV file.
        kwargs
            DuckDB CSV writer arguments. https://duckdb.org/docs/data/csv/overview.html#parameters

        """
        self._run_pre_execute_hooks(expr)
        query = self.compile(expr, params=params)
        args = [
            "FORMAT 'csv'",
            f"HEADER {int(header)}",
            *(f"{k.upper()} {v!r}" for k, v in kwargs.items()),
        ]
        copy_cmd = f"COPY ({query}) TO {str(path)!r} ({', '.join(args)})"
        with self._safe_raw_sql(copy_cmd):
            pass

    @util.experimental
    def to_geo(
        self,
        expr: ir.Table,
        /,
        path: str | Path,
        *,
        format: str,
        layer_creation_options: Mapping[str, Any] | None = None,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing `expr` to a geospatial output.

        Parameters
        ----------
        expr
            Ibis expression to execute and persist to geospatial output.
        path
            A string or Path to the desired output file location.
        format
            The format of the geospatial output. One of GDAL's supported vector formats.
            The list of vector formats is located here: https://gdal.org/en/latest/drivers/vector/index.html
        layer_creation_options
            A mapping of layer creation options.
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means no limit.
        kwargs
            Additional keyword arguments passed to the DuckDB `COPY` command.

        Examples
        --------
        >>> import os
        >>> import tempfile
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> from ibis import _

        Load some geospatial data

        >>> con = ibis.duckdb.connect()
        >>> zones = ibis.examples.zones.fetch(backend=con)
        >>> zones[["zone", "geom"]].head()
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ zone                                  ┃ geom                                 ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ string                                │ geospatial:geometry                  │
        ├───────────────────────────────────────┼──────────────────────────────────────┤
        │                                       │ <POLYGON ((933100.918 192536.086,    │
        │ Newark Airport                        │ 933091.011 192572.175, 933088.585    │
        │                                       │ 192604.9...>                         │
        │                                       │ <MULTIPOLYGON (((1033269.244         │
        │ Jamaica Bay                           │ 172126.008, 1033439.643 170883.946,  │
        │                                       │ 1033473.265...>                      │
        │                                       │ <POLYGON ((1026308.77 256767.698,    │
        │ Allerton/Pelham Gardens               │ 1026495.593 256638.616, 1026567.23   │
        │                                       │ 256589....>                          │
        │                                       │ <POLYGON ((992073.467 203714.076,    │
        │ Alphabet City                         │ 992068.667 203711.502, 992061.716    │
        │                                       │ 203711.7...>                         │
        │                                       │ <POLYGON ((935843.31 144283.336,     │
        │ Arden Heights                         │ 936046.565 144173.418, 936387.922    │
        │                                       │ 143967.75...>                        │
        └───────────────────────────────────────┴──────────────────────────────────────┘

        Write to a GeoJSON file

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     con.to_geo(
        ...         zones,
        ...         path=os.path.join(tmpdir, "zones.geojson"),
        ...         format="geojson",
        ...     )

        Write to a Shapefile

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     con.to_geo(
        ...         zones,
        ...         path=os.path.join(tmpdir, "zones.shp"),
        ...         format="ESRI Shapefile",
        ...     )
        """
        self._run_pre_execute_hooks(expr)
        query = self.compile(expr, params=params, limit=limit)

        args = ["FORMAT GDAL", f"DRIVER '{format}'"]

        if layer_creation_options := " ".join(
            f"{k.upper()}={v}" for k, v in (layer_creation_options or {}).items()
        ):
            args.append(f"LAYER_CREATION_OPTIONS '{layer_creation_options}'")

        args.extend(f"{k.upper()} {v!r}" for k, v in (kwargs or {}).items())

        copy_cmd = f"COPY ({query}) TO {str(path)!r} ({', '.join(args)})"

        with self._safe_raw_sql(copy_cmd):
            pass

    @util.experimental
    def to_json(
        self,
        expr: ir.Table,
        /,
        path: str | Path,
        *,
        compression: Literal["auto", "none", "gzip", "zstd"] = "auto",
        dateformat: str | None = None,
        timestampformat: str | None = None,
    ) -> None:
        """Write the results of `expr` to a json file of [{column -> value}, ...] objects.

        This method is eager and will execute the associated expression
        immediately.
        See https://duckdb.org/docs/sql/statements/copy.html#json-options
        for more info.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to Delta Lake table.
        path
            URLs such as S3 buckets are supported.
        compression
            Compression codec to use. One of "auto", "none", "gzip", "zstd".
        dateformat
            Date format string.
        timestampformat
            Timestamp format string.
        """
        opts = [f"COMPRESSION '{compression.upper()}'"]
        if dateformat:
            opts.append(f"DATEFORMAT '{dateformat}'")
        if timestampformat:
            opts.append(f"TIMESTAMPFORMAT '{timestampformat}'")

        options = f"FORMAT JSON, ARRAY TRUE, {', '.join(opts)}"

        self.raw_sql(f"COPY ({self.compile(expr)}) TO '{path!s}' ({options})")

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        with self._safe_raw_sql(f"DESCRIBE {query}") as cur:
            rows = cur.fetch_arrow_table()

        rows = rows.to_pydict()

        type_mapper = self.compiler.type_mapper
        return sch.Schema(
            {
                name: type_mapper.from_string(typ, nullable=null == "YES")
                for name, typ, null in zip(
                    rows["column_name"], rows["column_type"], rows["null"]
                )
            }
        )

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        data = op.data
        schema = op.schema

        try:
            obj = data.to_pyarrow_dataset(schema)
        except AttributeError:
            obj = data.to_pyarrow(schema)

        self.con.register(op.name, obj)

    def _register_udfs(self, expr: ir.Expr) -> None:
        con = self.con

        for udf_node in expr.op().find(ops.ScalarUDF):
            register_func = getattr(
                self, f"_register_{udf_node.__input_type__.name.lower()}_udf"
            )
            with contextlib.suppress(duckdb.InvalidInputException):
                con.remove_function(udf_node.__class__.__name__)

            registration_func = register_func(udf_node)
            if registration_func is not None:
                registration_func(con)

    def _register_udf(self, udf_node: ops.ScalarUDF):
        type_mapper = self.compiler.type_mapper
        input_types = [
            type_mapper.to_string(param.annotation.pattern.dtype)
            for param in udf_node.__signature__.parameters.values()
        ]

        def register_udf(con):
            return con.create_function(
                name=type(udf_node).__name__,
                function=udf_node.__func__,
                parameters=input_types,
                return_type=type_mapper.to_string(udf_node.dtype),
                type=_UDF_INPUT_TYPE_MAPPING[udf_node.__input_type__],
                **udf_node.__config__,
            )

        return register_udf

    _register_python_udf = _register_udf
    _register_pyarrow_udf = _register_udf

    def _get_temp_view_definition(self, name: str, definition: str) -> str:
        return sge.Create(
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            kind="VIEW",
            expression=definition,
            replace=True,
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        )

    def _create_temp_view(self, table_name, source):
        with self._safe_raw_sql(self._get_temp_view_definition(table_name, source)):
            pass


@lazy_singledispatch
def _read_in_memory(source: Any, table_name: str, _conn: Backend, **_: Any):
    raise NotImplementedError(
        f"The `{_conn.name}` backend currently does not support "
        f"reading data of {type(source)!r}"
    )


@_read_in_memory.register("polars.DataFrame")
@_read_in_memory.register("polars.LazyFrame")
@_read_in_memory.register("pyarrow.Table")
@_read_in_memory.register("pandas.DataFrame")
@_read_in_memory.register("pyarrow.dataset.Dataset")
def _default(source, table_name, _conn, **_: Any):
    _conn.con.register(table_name, source)


@_read_in_memory.register("pyarrow.RecordBatchReader")
def _pyarrow_rbr(source, table_name, _conn, **_: Any):
    _conn.con.register(table_name, source)
    # Ensure the reader isn't marked as started, in case the name is
    # being overwritten.
    _conn._record_batch_readers_consumed[table_name] = False



from __future__ import annotations

import itertools

import pandas as pd
import pytest

import xorq.api as xo
import xorq.expr.selectors as s
from xorq.api import memtable
from xorq.expr.ml import _calculate_bounds
from xorq.tests.util import assert_frame_equal


pytest.importorskip("duckdb")
pytest.importorskip("psycopg")
pytest.importorskip("datafusion")


def test_train_test_splits_intersections():
    # This is testing the base case where a single float becomes ( 1-test_size , test_size ) proportion
    # Check counts and overlaps in train and test dataset
    N = 10000
    test_size = [0.1, 0.2, 0.7]

    # init table
    table = memtable([(i, "val") for i in range(N)], columns=["key1", "val"])
    results = list(
        xo.train_test_splits(
            table,
            unique_key="key1",
            test_sizes=test_size,
            num_buckets=N,
            random_seed=42,
        )
    )

    # make sure all splits mutually exclusive
    # These are all  a \ b  U  a intersect b  where b are the other splits
    element1 = results[0]
    complement1 = results[1].union(results[2])

    element2 = results[1]
    complement2 = results[0].union(results[2])

    element3 = results[2]
    complement3 = results[0].union(results[1])

    assert element1.union(complement1).join(table, how="anti").count().execute() == 0
    assert (
        element1.join(complement1, element1.key1 == complement1.key1).count().execute()
        == 0
    )

    assert element2.union(complement2).join(table, how="anti").count().execute() == 0
    assert (
        element2.join(complement2, element2.key1 == complement2.key1).count().execute()
        == 0
    )

    assert element3.union(complement3).join(table, how="anti").count().execute() == 0
    assert (
        element3.join(complement3, element3.key1 == complement3.key1).count().execute()
        == 0
    )


def test_train_test_split():
    # This is testing the base case where a single float becomes ( 1-test_size , test_size ) proportion
    # Check counts and overlaps in train and test dataset
    N = 100
    test_size = 0.25

    # init table
    table = memtable([(i, "val") for i in range(N)], columns=["key1", "val"])
    train_table, test_table = xo.train_test_splits(
        table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=42
    )

    # These values are for seed 42
    assert train_table.count().execute() == 75
    assert test_table.count().execute() == 25
    assert set(train_table.columns) == set(table.columns)
    assert set(test_table.columns) == set(table.columns)
    # make sure data unioned together is itself
    assert train_table.union(test_table).join(table, how="semi").count().execute() == N

    # Check reproducibility
    reproduced_train_table, reproduced_test_table = xo.train_test_splits(
        table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=42
    )
    assert_frame_equal(train_table.execute(), reproduced_train_table.execute())
    assert_frame_equal(test_table.execute(), reproduced_test_table.execute())

    # make sure it could generate different data with different random_seed
    different_train_table, different_test_table = xo.train_test_splits(
        table, unique_key="key1", test_sizes=test_size, num_buckets=N, random_seed=0
    )
    assert not train_table.execute().equals(different_train_table.execute())
    assert not test_table.execute().equals(different_test_table.execute())


@pytest.mark.parametrize(
    "con_name,unique_key",
    itertools.product(
        (None, "pandas", "xorq", "sqlite"),
        (
            "key1",
            ("key1",),
            ("key1", "key2"),
        ),
    ),
)
def test_train_test_split_parametrized(con_name, unique_key):
    N = 200
    test_size = 0.25
    tolerance = 0.01

    table = memtable(
        [(i, j, f"val-{i}-{j}") for i, j in ((i, i % N) for i in range(N**2))],
        columns=["key1", "key2", "val"],
    )
    if con_name is not None:
        table = table.into_backend(getattr(xo, con_name).connect())
    train_table, test_table = xo.train_test_splits(
        table,
        unique_key=unique_key,
        test_sizes=test_size,
        num_buckets=N,
        random_seed=42,
    )

    # FIXME: use single expr rather than two executions: assert (train_table.union(test_table).join(table, how="semi").count().execute() == N**2)
    # splits are mutually exclusive and joinly exhaustive
    (train, test) = (expr.execute() for expr in (train_table, test_table))
    assert train.merge(test).empty
    assert len(pd.concat((train, test)).drop_duplicates()) == N**2

    # target test size is roughly attained
    assert abs(len(test) / (test_size * N**2) - 1) < tolerance


def test_train_test_split_invalid_test_size():
    table = memtable({"key": [1, 2, 3]})
    with pytest.raises(ValueError, match="test size should be a float between 0 and 1"):
        xo.train_test_splits(table, unique_key="key", test_sizes=1.5)
    with pytest.raises(ValueError, match="test size should be a float between 0 and 1"):
        xo.train_test_splits(table, unique_key="key", test_sizes=-0.5)


def test_train_test_split_invalid_num_buckets_type():
    table = memtable({"key": [1, 2, 3]})
    with pytest.raises(ValueError, match="num_buckets must be an integer"):
        xo.train_test_splits(table, unique_key="key", test_sizes=0.5, num_buckets=10.5)


def test_train_test_split_invalid_num_buckets_value():
    table = memtable({"key": [1, 2, 3]})
    with pytest.raises(
        ValueError, match="num_buckets = 1 places all data into training set"
    ):
        xo.train_test_splits(table, unique_key="key", test_sizes=0.5, num_buckets=1)


def test_train_test_split_multiple_keys():
    data = {
        "key1": range(100),
        "key2": [chr(i % 26 + 65) for i in range(100)],  # A, B, C, ...
        "value": [i % 3 for i in range(100)],
    }
    table = memtable(data)
    train_table, test_table = xo.train_test_splits(
        table,
        unique_key=["key1", "key2"],
        test_sizes=0.25,
        num_buckets=10,
        random_seed=99,
    )
    assert train_table.union(test_table).join(table, how="anti").count().execute() == 0


def test_train_test_splits_deterministic_with_seed():
    table = memtable({"key": range(100), "value": range(100)})
    test_sizes = [0.4, 0.6]

    splits1 = list(
        xo.train_test_splits(
            table,
            test_sizes=test_sizes,
            unique_key="key",
            random_seed=123,
            num_buckets=10,
        )
    )
    splits2 = list(
        xo.train_test_splits(
            table,
            test_sizes=test_sizes,
            unique_key="key",
            random_seed=123,
            num_buckets=10,
        )
    )

    for s1, s2 in zip(splits1, splits2):
        assert_frame_equal(s1.execute(), s2.execute())


def test_train_test_splits_invalid_test_sizes():
    table = memtable({"key": [1, 2, 3], "value": [4, 5, 6]})
    with pytest.raises(ValueError, match="Test size must be float."):
        next(xo.train_test_splits(table, "key", ["a", "b"]))
    with pytest.raises(
        ValueError, match="test size should be a float between 0 and 1."
    ):
        next(xo.train_test_splits(table, [-0.1, 0.5], "key"))


def test_train_test_splits_must_sum_one():
    table = memtable({"key": [1, 2, 3], "value": [4, 5, 6]})
    with pytest.raises(ValueError, match="Test sizes must sum to 1"):
        next(xo.train_test_splits(table, [0.1, 0.5], "key"))


def test_train_test_splits_with_all_selector():
    N = 50
    table = memtable({"k": range(N), "v": range(N)})
    splits = list(
        xo.train_test_splits(
            table, test_sizes=0.2, unique_key=s.all(), num_buckets=N, random_seed=0
        )
    )
    assert len(splits) == 2
    train_table, test_table = splits
    assert train_table.union(test_table).join(table, how="anti").count().execute() == 0

    splits2 = list(
        xo.train_test_splits(
            table, test_sizes=0.2, unique_key=s.all(), num_buckets=N, random_seed=0
        )
    )
    for s1, s2 in zip(splits, splits2):
        assert_frame_equal(s1.execute(), s2.execute())


@pytest.mark.parametrize(
    "test_sizes",
    ((1 / n,) * n for n in range(2, 100, 5)),
)
def test_approx_sum(test_sizes):
    _calculate_bounds(test_sizes)


def test_calculate_bounds():
    test_sizes = (0.2, 0.3, 0.5)
    expected_bounds = ((0.0, 0.2), (0.2, 0.5), (0.5, 1.0))
    assert _calculate_bounds(test_sizes) == expected_bounds


def test_train_test_splits_num_buckets_gt_one():
    table = memtable({"key": range(100), "value": range(100)})
    test_sizes = [0.4, 0.6]
    with pytest.raises(
        ValueError,
        match="num_buckets = 1 places all data into training set. For any integer x  >=0 , x mod 1 = 0 . ",
    ):
        next(
            xo.train_test_splits(
                table, "key", test_sizes, random_seed=123, num_buckets=1
            )
        )


@pytest.mark.parametrize(
    "connect_method",
    (
        lambda: xo.connect(),
        lambda: xo.duckdb.connect(),
        lambda: xo.postgres.connect_env(),
        pytest.param(
            xo.datafusion.connect,
            marks=pytest.mark.xfail(
                reason="Compilation rule for 'Hash' operation is not define"
            ),
        ),
    ),
)
def test_train_test_splits_intersections_parameterized_pass(connect_method):
    # This is testing the base case where a single float becomes ( 1-test_size , test_size ) proportion
    # Check counts and overlaps in train and test dataset
    N = 10000
    test_size = [0.1, 0.2, 0.7]

    # create test table for backend
    test_df = pd.DataFrame([(i, "val") for i in range(N)], columns=["key1", "val"])
    con = connect_method()
    test_table_name = f"{con.name}_test_df"
    con.create_table(test_table_name, test_df, temp=con.name == "postgres")

    table = con.table(test_table_name)

    results = list(
        xo.train_test_splits(
            table,
            unique_key="key1",
            test_sizes=test_size,
            num_buckets=N,
            random_seed=42,
        )
    )

    # make sure all splits mutually exclusive
    # These are all  a \ b  U  a intersect b  where b are the other splits
    element1 = results[0]
    complement1 = results[1].union(results[2])

    element2 = results[1]
    complement2 = results[0].union(results[2])

    element3 = results[2]
    complement3 = results[0].union(results[1])

    assert element1.union(complement1).join(table, how="anti").count().execute() == 0
    assert (
        element1.join(complement1, element1.key1 == complement1.key1).count().execute()
        == 0
    )

    assert element2.union(complement2).join(table, how="anti").count().execute() == 0
    assert (
        element2.join(complement2, element2.key1 == complement2.key1).count().execute()
        == 0
    )

    assert element3.union(complement3).join(table, how="anti").count().execute() == 0
    assert (
        element3.join(complement3, element3.key1 == complement3.key1).count().execute()
        == 0
    )
    con.drop_table(test_table_name)


@pytest.mark.parametrize(
    "connect_method",
    (
        lambda: xo.connect(),
        lambda: xo.duckdb.connect(),
        lambda: xo.postgres.connect_env(),
        pytest.param(
            lambda: xo.datafusion.connect(),
            marks=pytest.mark.xfail(
                reason="Compilation rule for 'Hash' operation is not define"
            ),
        ),
    ),
)
@pytest.mark.parametrize("n", (2, 8, 32))
@pytest.mark.parametrize("name", ("split", "other"))
def test_calc_split_column(connect_method, n, name):
    N = 10000
    test_sizes = (1 / n,) * n
    unique_key = "key1"

    # create test table for backend
    test_df = pd.DataFrame([(i, "val") for i in range(N)], columns=[unique_key, "val"])
    con = connect_method()
    test_table_name = f"{con.name}_test_df"
    con.create_table(test_table_name, test_df, temp=con.name == "postgres")

    table = con.table(test_table_name)
    expr = (
        table.mutate(
            xo.calc_split_column(
                table,
                unique_key=unique_key,
                test_sizes=test_sizes,
                random_seed=42,
                name=name,
            )
        )[name]
        .value_counts()
        .order_by(xo.asc(name))
    )
    df = xo.execute(expr)
    assert tuple(df[name].values) == tuple(range(n))
    assert df[f"{name}_count"].sum() == N



from __future__ import annotations

import inspect
import re
import sys
from abc import abstractmethod
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, Optional, TypeVar, get_args, get_origin
from typing import get_type_hints as _get_type_hints

from xorq.vendor.ibis.common.bases import Abstract
from xorq.vendor.ibis.common.caching import memoize


if TYPE_CHECKING:
    from typing_extensions import Self

from types import UnionType
from typing import TypeAlias


# Keep this alias in sync with unittest.case._ClassInfo
_ClassInfo: TypeAlias = type | UnionType | tuple["_ClassInfo", ...]


T = TypeVar("T")
U = TypeVar("U")

Namespace = dict[str, Any]
VarTuple = tuple[T, ...]


@memoize
def get_type_hints(
    obj: Any,
    include_extras: bool = True,
    include_properties: bool = False,
) -> dict[str, Any]:
    """Get type hints for a callable or class.

    Extension of typing.get_type_hints that supports getting type hints for
    class properties.

    Parameters
    ----------
    obj
        Callable or class to get type hints for.
    include_extras
        Whether to include extra type hints such as Annotated.
    include_properties
        Whether to include type hints for class properties.

    Returns
    -------
    Mapping of parameter or attribute name to type hint.

    """
    try:
        hints = _get_type_hints(obj, include_extras=include_extras)
    except TypeError:
        return {}

    if include_properties:
        for name in dir(obj):
            attr = getattr(obj, name)
            if isinstance(attr, property):
                annots = _get_type_hints(attr.fget, include_extras=include_extras)
                if return_annot := annots.get("return"):
                    hints[name] = return_annot

    return hints


@memoize
def get_type_params(obj: Any) -> dict[str, type]:
    """Get type parameters for a generic class.

    Parameters
    ----------
    obj
        Generic class to get type parameters for.

    Returns
    -------
    Mapping of type parameter name to type.

    Examples
    --------
    >>> from typing import Dict, List
    >>> class MyList(List[T]): ...
    >>> get_type_params(MyList[int])
    {'T': <class 'int'>}
    >>> class MyDict(Dict[T, U]): ...
    >>> get_type_params(MyDict[int, str])
    {'T': <class 'int'>, 'U': <class 'str'>}

    """
    args = get_args(obj)
    origin = get_origin(obj) or obj
    bases = getattr(origin, "__orig_bases__", ())
    params = getattr(origin, "__parameters__", ())

    result = {}
    for base in bases:
        result.update(get_type_params(base))

    param_names = (p.__name__ for p in params)
    result.update(zip(param_names, args))

    return result


@memoize
def get_bound_typevars(obj: Any) -> dict[TypeVar, tuple[str, type]]:
    """Get type variables bound to concrete types for a generic class.

    Parameters
    ----------
    obj
        Generic class to get type variables for.

    Returns
    -------
    Mapping of type variable to attribute name and type.

    Examples
    --------
    >>> from typing import Generic
    >>> class MyStruct(Generic[T, U]):
    ...     a: T
    ...     b: U
    >>> get_bound_typevars(MyStruct[int, str])
    {~T: ('a', <class 'int'>), ~U: ('b', <class 'str'>)}
    >>>
    >>> class MyStruct(Generic[T, U]):
    ...     a: T
    ...
    ...     @property
    ...     def myprop(self) -> U: ...
    >>> get_bound_typevars(MyStruct[float, bytes])
    {~T: ('a', <class 'float'>), ~U: ('myprop', <class 'bytes'>)}

    """
    origin = get_origin(obj) or obj
    hints = get_type_hints(origin, include_properties=True)
    params = get_type_params(obj)

    result = {}
    for attr, typ in hints.items():
        if isinstance(typ, TypeVar):
            result[typ] = (attr, params[typ.__name__])
    return result


def evaluate_annotations(
    annots: dict[str, str],
    module_name: str,
    class_name: Optional[str] = None,
    best_effort: bool = False,
) -> dict[str, Any]:
    """Evaluate type annotations that are strings.

    Parameters
    ----------
    annots
        Type annotations to evaluate.
    module_name
        The name of the module that the annotations are defined in, hence
        providing global scope.
    class_name
        The name of the class that the annotations are defined in, hence
        providing Self type.
    best_effort
        Whether to ignore errors when evaluating type annotations.

    Returns
    -------
    Actual type hints.

    Examples
    --------
    >>> annots = {"a": "dict[str, float]", "b": "int"}
    >>> evaluate_annotations(annots, __name__)
    {'a': dict[str, float], 'b': <class 'int'>}

    """
    module = sys.modules.get(module_name, None)
    globalns = getattr(module, "__dict__", None)
    if class_name is None:
        localns = None
    else:
        localns = {"Self": f"{module_name}.{class_name}"}

    result = {}
    for k, v in annots.items():
        if isinstance(v, str):
            try:
                v = eval(v, globalns, localns)  # noqa: S307
            except NameError:
                if not best_effort:
                    raise
        result[k] = v

    return result


def format_typehint(typ: Any) -> str:
    if isinstance(typ, type):
        return typ.__name__
    elif isinstance(typ, TypeVar):
        if typ.__bound__ is None:
            return str(typ)
        else:
            return format_typehint(typ.__bound__)
    else:
        # remove the module name from the typehint, including generics
        return re.sub(r"(\w+\.)+", "", str(typ))


class DefaultTypeVars:
    """Enable using default type variables in generic classes (PEP-0696)."""

    __slots__ = ()

    def __class_getitem__(cls, params):
        params = params if isinstance(params, tuple) else (params,)
        pairs = zip_longest(params, cls.__parameters__)
        params = tuple(p.__default__ if t is None else t for t, p in pairs)
        return super().__class_getitem__(params)


class Sentinel(type):
    """Create type-annotable unique objects."""

    def __new__(cls, name, bases, namespace, **kwargs):
        if bases:
            raise TypeError("Sentinels cannot be subclassed")
        return super().__new__(cls, name, bases, namespace, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise TypeError("Sentinels are not constructible")


class CoercionError(Exception): ...


class Coercible(Abstract):
    """Protocol for defining coercible types.

    Coercible types define a special `__coerce__` method that accepts an object
    with an instance of the type. Used in conjunction with the `coerced_to``
    pattern to coerce arguments to a specific type.
    """

    @classmethod
    @abstractmethod
    def __coerce__(cls, value: Any, **kwargs: Any) -> Self: ...


def get_defining_frame(obj):
    """Locate the outermost frame where `obj` is defined."""
    for frame_info in inspect.stack()[::-1]:
        for var in frame_info.frame.f_locals.values():
            if obj is var:
                return frame_info.frame
    raise ValueError(f"No defining frame found for {obj}")


def get_defining_scope(obj, types=None):
    """Get variables in the scope where `expr` is first defined."""
    frame = get_defining_frame(obj)
    scope = {**frame.f_globals, **frame.f_locals}
    if types is not None:
        scope = {k: v for k, v in scope.items() if isinstance(v, types)}
    return scope



"""Temporal operations."""

from __future__ import annotations

import operator
from typing import Annotated, Optional

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.common.patterns import As, Attrs
from xorq.vendor.ibis.common.temporal import (
    DateUnit,
    IntervalUnit,
    TimestampUnit,
    TimeUnit,
)
from xorq.vendor.ibis.expr.operations.core import Binary, Scalar, Unary, Value
from xorq.vendor.ibis.expr.operations.logical import Between


@public
class TimestampTruncate(Value):
    """Truncate a timestamp to a specified unit."""

    arg: Value[dt.Timestamp]
    unit: IntervalUnit

    shape = rlz.shape_like("arg")
    dtype = dt.timestamp


@public
class DateTruncate(Value):
    """Truncate a date to a specified unit."""

    arg: Value[dt.Date]
    unit: DateUnit

    shape = rlz.shape_like("arg")
    dtype = dt.date


@public
class TimeTruncate(Value):
    """Truncate a time to a specified unit."""

    arg: Value[dt.Time]
    unit: TimeUnit

    shape = rlz.shape_like("arg")
    dtype = dt.time


@public
class TimestampBucket(Value):
    """Bucketize a timestamp to a specified interval."""

    arg: Value[dt.Timestamp]
    interval: Scalar[dt.Interval]
    offset: Optional[Scalar[dt.Interval]] = None

    shape = rlz.shape_like("arg")
    dtype = dt.timestamp


@public
class Strftime(Value):
    """Format a temporal value as a string."""

    arg: Value[dt.Temporal]
    format_str: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.string


@public
class StringToTimestamp(Value):
    """Convert a string to a timestamp."""

    arg: Value[dt.String]
    format_str: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.Timestamp(timezone="UTC")


@public
class StringToDate(Value):
    """Convert a string to a date."""

    arg: Value[dt.String]
    format_str: Value[dt.String]

    shape = rlz.shape_like("args")
    dtype = dt.date


@public
class ExtractTemporalField(Unary):
    """Extract a field from a temporal value."""

    arg: Value[dt.Temporal]
    dtype = dt.int32


@public
class ExtractDateField(ExtractTemporalField):
    """Extract a field from a date."""

    arg: Value[dt.Date | dt.Timestamp]


@public
class ExtractTimeField(ExtractTemporalField):
    """Extract a field from a time."""

    arg: Value[dt.Time | dt.Timestamp]


@public
class ExtractYear(ExtractDateField):
    """Extract the year from a date or timestamp."""


@public
class ExtractIsoYear(ExtractDateField):
    """Extract the ISO year from a date or timestamp."""


@public
class ExtractMonth(ExtractDateField):
    """Extract the month from a date or timestamp."""


@public
class ExtractDay(ExtractDateField):
    """Extract the day from a date or timestamp."""


@public
class ExtractDayOfYear(ExtractDateField):
    """Extract the day of the year from a date or timestamp."""


@public
class ExtractQuarter(ExtractDateField):
    """Extract the quarter from a date or timestamp."""


@public
class ExtractEpochSeconds(ExtractDateField):
    """Extract seconds since the UNIX epoch from a date or timestamp."""


@public
class ExtractWeekOfYear(ExtractDateField):
    """Extract the week of the year from a date or timestamp."""


@public
class ExtractHour(ExtractTimeField):
    """Extract the hour from a time or timestamp."""


@public
class ExtractMinute(ExtractTimeField):
    """Extract the minute from a time or timestamp."""


@public
class ExtractSecond(ExtractTimeField):
    """Extract the second from a time or timestamp."""


@public
class ExtractMillisecond(ExtractTimeField):
    """Extract milliseconds from a time or timestamp."""


@public
class ExtractMicrosecond(ExtractTimeField):
    """Extract microseconds from a time or timestamp."""


@public
class DayOfWeekIndex(Unary):
    """Extract the index of the day of the week from a date or timestamp."""

    arg: Value[dt.Date | dt.Timestamp]

    dtype = dt.int16


@public
class DayOfWeekName(Unary):
    """Extract the name of the day of the week from a date or timestamp."""

    arg: Value[dt.Date | dt.Timestamp]

    dtype = dt.string


@public
class Time(Unary):
    """Extract the time from a timestamp."""

    dtype = dt.time


@public
class Date(Unary):
    """Extract the date from a timestamp."""

    dtype = dt.date


@public
class DateFromYMD(Value):
    """Construct a date from year, month, and day."""

    year: Value[dt.Integer]
    month: Value[dt.Integer]
    day: Value[dt.Integer]

    dtype = dt.date
    shape = rlz.shape_like("args")


@public
class TimeFromHMS(Value):
    """Construct a time from hours, minutes, and seconds."""

    hours: Value[dt.Integer]
    minutes: Value[dt.Integer]
    seconds: Value[dt.Integer]

    dtype = dt.time
    shape = rlz.shape_like("args")


@public
class TimestampFromYMDHMS(Value):
    """Construct a timestamp from components."""

    year: Value[dt.Integer]
    month: Value[dt.Integer]
    day: Value[dt.Integer]
    hours: Value[dt.Integer]
    minutes: Value[dt.Integer]
    seconds: Value[dt.Integer]

    dtype = dt.timestamp
    shape = rlz.shape_like("args")


@public
class TimestampFromUNIX(Value):
    """Construct a timestamp from a UNIX timestamp."""

    arg: Value
    unit: TimestampUnit

    dtype = dt.timestamp
    shape = rlz.shape_like("arg")


TimeInterval = Annotated[dt.Interval, Attrs(unit=As(TimeUnit))]
DateInterval = Annotated[dt.Interval, Attrs(unit=As(DateUnit))]


@public
class DateAdd(Binary):
    """Add an interval to a date."""

    left: Value[dt.Date]
    right: Value[DateInterval]

    dtype = rlz.dtype_like("left")


@public
class DateSub(Binary):
    """Subtract an interval from a date."""

    left: Value[dt.Date]
    right: Value[DateInterval]

    dtype = rlz.dtype_like("left")


@public
class DateDiff(Binary):
    """Compute the difference between two dates."""

    left: Value[dt.Date]
    right: Value[dt.Date]

    dtype = dt.Interval("D")


@public
class TimeAdd(Binary):
    """Add an interval to a time."""

    left: Value[dt.Time]
    right: Value[TimeInterval]

    dtype = rlz.dtype_like("left")


@public
class TimeSub(Binary):
    """Subtract an interval from a time."""

    left: Value[dt.Time]
    right: Value[TimeInterval]

    dtype = rlz.dtype_like("left")


@public
class TimeDiff(Binary):
    """Compute the difference between two times."""

    left: Value[dt.Time]
    right: Value[dt.Time]

    dtype = dt.Interval("s")


@public
class TimestampAdd(Binary):
    """Add an interval to a timestamp."""

    left: Value[dt.Timestamp]
    right: Value[dt.Interval]

    dtype = rlz.dtype_like("left")


@public
class TimestampSub(Binary):
    """Subtract an interval from a timestamp."""

    left: Value[dt.Timestamp]
    right: Value[dt.Interval]

    dtype = rlz.dtype_like("left")


@public
class TimestampDiff(Binary):
    """Compute the difference between two timestamps."""

    left: Value[dt.Timestamp]
    right: Value[dt.Timestamp]

    dtype = dt.Interval("s")


@public
class IntervalBinary(Binary):
    """Base class for interval binary operations."""

    @attribute
    def dtype(self):
        interval_unit_args = [
            arg.dtype.unit for arg in (self.left, self.right) if arg.dtype.is_interval()
        ]
        unit = rlz._promote_interval_resolution(interval_unit_args)

        return self.left.dtype.copy(unit=unit)


@public
class IntervalAdd(IntervalBinary):
    """Add two intervals."""

    left: Value[dt.Interval]
    right: Value[dt.Interval]
    op = operator.add


@public
class IntervalSubtract(IntervalBinary):
    """Subtract one interval from another."""

    left: Value[dt.Interval]
    right: Value[dt.Interval]
    op = operator.sub


@public
class IntervalMultiply(IntervalBinary):
    """Multiply an interval by a scalar."""

    left: Value[dt.Interval]
    right: Value[dt.Numeric | dt.Boolean]
    op = operator.mul


@public
class IntervalFloorDivide(IntervalBinary):
    """Divide an interval by a scalar, rounding down."""

    left: Value[dt.Interval]
    right: Value[dt.Numeric | dt.Boolean]
    op = operator.floordiv


@public
class IntervalFromInteger(Value):
    """Construct an interval from an integer."""

    arg: Value[dt.Integer]
    unit: IntervalUnit

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self):
        return dt.Interval(self.unit)

    @property
    def resolution(self):
        return self.dtype.resolution


@public
class BetweenTime(Between):
    """Check if a time is between two bounds."""

    arg: Value[dt.Time | dt.Timestamp]
    lower_bound: Value[dt.Time | dt.String]
    upper_bound: Value[dt.Time | dt.String]


class TemporalDelta(Value):
    """Base class for temporal delta operations."""

    part: Value[dt.String]
    shape = rlz.shape_like("args")
    dtype = dt.int64


@public
class TimeDelta(TemporalDelta):
    """Compute the difference between two times as integer number of requested units."""

    left: Value[dt.Time]
    right: Value[dt.Time]


@public
class DateDelta(TemporalDelta):
    """Compute the difference between two dates as integer number of requested units."""

    left: Value[dt.Date]
    right: Value[dt.Date]


@public
class TimestampDelta(TemporalDelta):
    """Compute the difference between two timestamps as integer number of requested units."""

    left: Value[dt.Timestamp]
    right: Value[dt.Timestamp]


public(ExtractTimestampField=ExtractTemporalField)
